#------------------------------------------------------------------------------
# Copyright (c) 2020, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# buffer.pyx
#
# Cython file defining the low-level network buffer read and write classes and
# methods for reading and writing low-level data from those buffers (embedded
# in thin_impl.pyx).
#------------------------------------------------------------------------------

cdef enum:
    PACKET_HEADER_SIZE = 8
    CHUNKED_BYTES_CHUNK_SIZE = 65536

cdef struct BytesChunk:
    char_type *ptr
    uint32_t length
    uint32_t allocated_length

cdef struct Rowid:
    uint32_t rba
    uint16_t partition_id
    uint32_t block_num
    uint16_t slot_num

@cython.final
@cython.freelist(20)
cdef class Packet:

    cdef:
        uint32_t packet_size
        uint8_t packet_type
        uint8_t packet_flags
        bytes buf

    cdef inline bint has_end_of_response(self):
        """
        Returns a boolean indicating if the end of request byte is found at the
        end of the packet.
        """
        cdef:
            uint16_t flags
            char *ptr
        ptr = cpython.PyBytes_AS_STRING(self.buf)
        flags = decode_uint16be(<const char_type*> &ptr[PACKET_HEADER_SIZE])
        if flags & TNS_DATA_FLAGS_END_OF_RESPONSE or flags & TNS_DATA_FLAGS_EOF:
            return True
        if self.packet_size == PACKET_HEADER_SIZE + 3 \
                and ptr[PACKET_HEADER_SIZE + 2] == TNS_MSG_TYPE_END_OF_RESPONSE:
            return True
        return False


@cython.final
cdef class ChunkedBytesBuffer:

    cdef:
        uint32_t _num_chunks
        uint32_t _allocated_chunks
        BytesChunk *_chunks

    def __dealloc__(self):
        cdef uint32_t i
        for i in range(self._allocated_chunks):
            if self._chunks[i].ptr is not NULL:
                cpython.PyMem_Free(self._chunks[i].ptr)
                self._chunks[i].ptr = NULL
        if self._chunks is not NULL:
            cpython.PyMem_Free(self._chunks)
            self._chunks = NULL

    cdef int _allocate_chunks(self) except -1:
        """
        Allocates a new set of chunks and copies data from the original set of
        chunks if needed.
        """
        cdef:
            BytesChunk *chunks
            uint32_t allocated_chunks
        allocated_chunks = self._allocated_chunks + 8
        chunks = <BytesChunk*> \
                cpython.PyMem_Malloc(sizeof(BytesChunk) * allocated_chunks)
        memset(chunks, 0, sizeof(BytesChunk) * allocated_chunks)
        if self._num_chunks > 0:
            memcpy(chunks, self._chunks, sizeof(BytesChunk) * self._num_chunks)
            cpython.PyMem_Free(self._chunks)
        self._chunks = chunks
        self._allocated_chunks = allocated_chunks

    cdef BytesChunk* _get_chunk(self, uint32_t num_bytes) except NULL:
        """
        Return the chunk that can be used to write the number of bytes
        requested.
        """
        cdef:
            uint32_t num_allocated_bytes
            BytesChunk *chunk
        if self._num_chunks > 0:
            chunk = &self._chunks[self._num_chunks - 1]
            if chunk.allocated_length >= chunk.length + num_bytes:
                return chunk
        if self._num_chunks >= self._allocated_chunks:
            self._allocate_chunks()
        self._num_chunks += 1
        chunk = &self._chunks[self._num_chunks - 1]
        chunk.length = 0
        if chunk.allocated_length < num_bytes:
            num_allocated_bytes = self._get_chunk_size(num_bytes)
            if chunk.ptr:
                cpython.PyMem_Free(chunk.ptr)
            chunk.ptr = <char_type*> cpython.PyMem_Malloc(num_allocated_bytes)
            chunk.allocated_length = num_allocated_bytes
        return chunk

    cdef inline uint32_t _get_chunk_size(self, uint32_t size):
        """
        Returns the size to allocate aligned on a 64K boundary.
        """
        return (size + CHUNKED_BYTES_CHUNK_SIZE - 1) & \
               ~(CHUNKED_BYTES_CHUNK_SIZE - 1)

    cdef char_type* end_chunked_read(self) except NULL:
        """
        Called when a chunked read has ended. Since a chunked read is never
        started until at least some bytes are being read, it is assumed that at
        least one chunk is in use. If one chunk is in use, those bytes are
        returned directly, but if more than one chunk is in use, the first
        chunk is resized to include all of the bytes in a contiguous section of
        memory first.
        """
        cdef:
            uint32_t i, num_allocated_bytes, total_num_bytes = 0, pos = 0
            char_type *ptr
        if self._num_chunks > 1:
            for i in range(self._num_chunks):
                total_num_bytes += self._chunks[i].length
            num_allocated_bytes = self._get_chunk_size(total_num_bytes)
            ptr = <char_type*> cpython.PyMem_Malloc(num_allocated_bytes)
            for i in range(self._num_chunks):
                memcpy(&ptr[pos], self._chunks[i].ptr, self._chunks[i].length)
                pos += self._chunks[i].length
                cpython.PyMem_Free(self._chunks[i].ptr)
                self._chunks[i].ptr = NULL
                self._chunks[i].allocated_length = 0
                self._chunks[i].length = 0
            self._num_chunks = 1
            self._chunks[0].ptr = ptr
            self._chunks[0].length = total_num_bytes
            self._chunks[0].allocated_length = num_allocated_bytes
        return self._chunks[0].ptr

    cdef char_type* get_chunk_ptr(self, uint32_t size_required) except NULL:
        """
        Called when memory is required for a chunked read.
        """
        cdef:
            BytesChunk *chunk
            char_type *ptr
        chunk = self._get_chunk(size_required)
        ptr = &chunk.ptr[chunk.length]
        chunk.length += size_required
        return ptr

    cdef inline void start_chunked_read(self):
        """
        Called when a chunked read is started and simply indicates that no
        chunks are in use. The memory is retained in order to reduce the
        overhead in freeing and reallocating memory for each chunked read.
        """
        self._num_chunks = 0


@cython.final
cdef class ReadBuffer(Buffer):

    cdef:
        ssize_t _saved_packet_pos, _next_packet_pos, _saved_pos
        ChunkedBytesBuffer _chunked_bytes_buf
        const char_type _split_data[255]
        uint32_t _pending_error_num
        Packet _current_packet
        Transport _transport
        list _saved_packets
        Capabilities _caps
        bint _check_request_boundary
        bint _in_pipeline
        object _waiter
        object _loop

    def __cinit__(self, Transport transport, Capabilities caps):
        self._transport = transport
        self._caps = caps
        self._chunked_bytes_buf = ChunkedBytesBuffer()

    cdef int _check_connected(self):
        """
        Checks to see if the transport is connected and throws the appropriate
        exception if not.
        """
        if self._pending_error_num not in (
            0, TNS_ERR_SESSION_SHUTDOWN, TNS_ERR_INBAND_MESSAGE
        ):
            if self._transport is not None:
                self._transport.disconnect()
                self._transport = None
            if self._pending_error_num == TNS_ERR_EXCEEDED_IDLE_TIME:
                errors._raise_err(errors.ERR_EXCEEDED_IDLE_TIME)
            else:
                errors._raise_err(errors.ERR_UNSUPPORTED_INBAND_NOTIFICATION,
                                  err_num=self._pending_error_num)
        elif self._transport is None:
            if self._pending_error_num == TNS_ERR_SESSION_SHUTDOWN:
                errors._raise_err(errors.ERR_CONNECTION_CLOSED)
            errors._raise_err(errors.ERR_NOT_CONNECTED)

    cdef int _get_int_length_and_sign(self, uint8_t *length,
                                      bint *is_negative,
                                      uint8_t max_length) except -1:
        """
        Returns the length of an integer sent on the wire. A check is also made
        to ensure the integer does not exceed the maximum length. If the
        is_negative pointer is NULL, negative integers will result in an
        exception being raised.
        """
        cdef const char_type *ptr = self._get_raw(1)
        if ptr[0] & 0x80:
            if is_negative == NULL:
                errors._raise_err(errors.ERR_UNEXPECTED_NEGATIVE_INTEGER)
            is_negative[0] = True
            length[0] = ptr[0] & 0x7f
        else:
            if is_negative != NULL:
                is_negative[0] = False
            length[0] = ptr[0]
        if length[0] > max_length:
            errors._raise_err(errors.ERR_INTEGER_TOO_LARGE, length=length[0],
                              max_length=max_length)

    cdef const char_type* _get_raw(self, ssize_t num_bytes,
                                   bint in_chunked_read=False) except NULL:
        """
        Returns a pointer to a buffer containing the requested number of bytes.
        This may be split across multiple packets in which case a chunked bytes
        buffer is used.
        """
        cdef:
            ssize_t num_bytes_left, num_bytes_split, max_split_data
            const char_type *source_ptr
            char_type *dest_ptr

        # if no bytes are left in the buffer, a new packet needs to be fetched
        # before anything else can take place
        if self._pos == self._size:
            self.wait_for_packets_sync(check_marker=True)

        # if there is enough room in the buffer to satisfy the number of bytes
        # requested, return a pointer to the current location and advance the
        # offset the required number of bytes
        source_ptr = &self._data[self._pos]
        num_bytes_left = self._size - self._pos
        if num_bytes <= num_bytes_left:
            if in_chunked_read:
                dest_ptr = self._chunked_bytes_buf.get_chunk_ptr(num_bytes)
                memcpy(dest_ptr, source_ptr, num_bytes)
            self._pos += num_bytes
            return source_ptr

        # the requested bytes are split across multiple packets; if a chunked
        # read is in progress, a chunk is acquired that will accommodate the
        # remainder of the bytes in the current packet; otherwise, the split
        # buffer will be used instead (after first checking to see if there is
        # sufficient room available within it)
        if in_chunked_read:
            dest_ptr = self._chunked_bytes_buf.get_chunk_ptr(num_bytes_left)
        else:
            max_split_data = sizeof(self._split_data)
            if max_split_data < num_bytes:
                errors._raise_err(errors.ERR_BUFFER_LENGTH_INSUFFICIENT,
                                  actual_buffer_len=max_split_data,
                                  required_buffer_len=num_bytes)
            dest_ptr = <char_type*> self._split_data
        memcpy(dest_ptr, source_ptr, num_bytes_left)

        # acquire packets until the requested number of bytes is satisfied
        num_bytes -= num_bytes_left
        while num_bytes > 0:

            # advance to next packet
            self.wait_for_packets_sync(check_marker=True)

            # copy data into the chunked buffer or split buffer, as appropriate
            source_ptr = &self._data[self._pos]
            num_bytes_split = min(num_bytes, self._size - self._pos)
            if in_chunked_read:
                dest_ptr = \
                        self._chunked_bytes_buf.get_chunk_ptr(num_bytes_split)
            else:
                dest_ptr = <char_type*> &self._split_data[num_bytes_left]
            memcpy(dest_ptr, source_ptr, num_bytes_split)
            self._pos += num_bytes_split
            num_bytes -= num_bytes_split

        # return the split buffer unconditionally; if performing a chunked read
        # the return value is ignored anyway
        return self._split_data

    cdef int _process_control_packet(self, Packet packet) except -1:
        """
        Processes a control packet.
        """
        cdef:
            uint16_t control_type
            Buffer buf
        buf = Buffer.__new__(Buffer)
        buf._populate_from_bytes(packet.buf)
        buf.skip_raw_bytes(8)               # skip packet header
        buf.read_uint16be(&control_type)
        if control_type == TNS_CONTROL_TYPE_RESET_OOB:
            self._caps.supports_oob = False
        elif control_type == TNS_CONTROL_TYPE_INBAND_NOTIFICATION:
            buf.skip_raw_bytes(4)           # skip first integer
            buf.read_uint32be(&self._pending_error_num)

    cdef int _process_packet(self, Packet packet,
                             bint *notify_waiter,
                             bint check_connected) except -1:
        """
        Processes a packet. If the packet is a control packet it is processed
        immediately and discarded; if the packet is a marker packet and a
        pipeline is being processed, the packet is discarded; otherwise, it is
        added to the list of saved packets for this response. If the protocol
        supports sending the end of request notification we wait for that
        message type to be returned at the end of the packet.
        """
        if packet.packet_type == TNS_PACKET_TYPE_CONTROL:
            self._process_control_packet(packet)
            notify_waiter[0] = False
            if check_connected:
                self._check_connected()
        elif self._in_pipeline \
                and packet.packet_type == TNS_PACKET_TYPE_MARKER:
            notify_waiter[0] = False
        else:
            self._saved_packets.append(packet)
            notify_waiter[0] = \
                    packet.packet_type != TNS_PACKET_TYPE_DATA \
                    or not self._check_request_boundary \
                    or packet.has_end_of_response()

    cdef int _read_raw_bytes_and_length(self, const char_type **ptr,
                                        ssize_t *num_bytes) except -1:
        """
        Helper function that processes the length. If the length is defined as
        TNS_LONG_LENGTH_INDICATOR, a chunked read is performed.
        """
        cdef uint32_t temp_num_bytes
        if num_bytes[0] != TNS_LONG_LENGTH_INDICATOR:
            return Buffer._read_raw_bytes_and_length(self, ptr, num_bytes)
        self._chunked_bytes_buf.start_chunked_read()
        num_bytes[0] = 0
        while True:
            self.read_ub4(&temp_num_bytes)
            if temp_num_bytes == 0:
                break
            num_bytes[0] += temp_num_bytes
            self._get_raw(temp_num_bytes, in_chunked_read=True)
        ptr[0] = self._chunked_bytes_buf.end_chunked_read()

    cdef int _start_packet(self) except -1:
        """
        Starts a packet. This prepares the current packet for processing.
        """
        cdef uint16_t data_flags
        self._current_packet = self._saved_packets[self._next_packet_pos]
        self._next_packet_pos += 1
        self._populate_from_bytes(self._current_packet.buf)
        self._pos = PACKET_HEADER_SIZE
        if self._current_packet.packet_type == TNS_PACKET_TYPE_DATA:
            self.read_uint16be(&data_flags)
            if data_flags == TNS_DATA_FLAGS_EOF:
                self._pending_error_num = TNS_ERR_SESSION_SHUTDOWN

    async def discard_pipeline_responses(self, ssize_t num_responses):
        """
        Discards the specified number of responses after the pipeline has
        encountered an exception.
        """
        while num_responses > 0:
            if not self.has_response():
                await self.wait_for_response_async()
            while True:
                self._start_packet()
                if self._current_packet.has_end_of_response():
                    break
            num_responses -= 1
        self.reset_packets()

    cdef int notify_packet_received(self) except -1:
        """
        Notify the registered waiter that a packet has been received. This is
        only used by the asyncio implementation.
        """
        if self._waiter is not None:
            self._waiter.set_result(None)
            self._waiter = None

    cdef ThinDbObjectImpl read_dbobject(self, BaseDbObjectTypeImpl typ_impl):
        """
        Read a database object from the buffer and return a DbObject object
        containing it.
        it.
        """
        cdef:
            bytes oid = None, toid = None
            ThinDbObjectImpl obj_impl
            uint32_t num_bytes
        self.read_ub4(&num_bytes)
        if num_bytes > 0:                   # type OID
            toid = self.read_bytes()
        self.read_ub4(&num_bytes)
        if num_bytes > 0:                   # OID
            oid = self.read_bytes()
        self.read_ub4(&num_bytes)
        if num_bytes > 0:                   # snapshot
            self.read_bytes()
        self.skip_ub2()                     # version
        self.read_ub4(&num_bytes)           # length of data
        self.skip_ub2()                     # flags
        if num_bytes > 0:
            obj_impl = ThinDbObjectImpl.__new__(ThinDbObjectImpl)
            obj_impl.type = typ_impl
            obj_impl.toid = toid
            obj_impl.oid = oid
            obj_impl.packed_data = self.read_bytes()
            return obj_impl

    cdef object read_oson(self):
        """
        Read an OSON value from the buffer and return the converted value. OSON
        is sent as a LOB value with all of the data prefetched. Since the LOB
        locator is not required it is simply discarded.
        it.
        """
        cdef:
            OsonDecoder decoder
            uint32_t num_bytes
            bytes data
        self.read_ub4(&num_bytes)
        if num_bytes > 0:
            self.skip_ub8()             # size (unused)
            self.skip_ub4()             # chunk size (unused)
            data = self.read_bytes()
            self.read_bytes()           # LOB locator (unused)
            decoder = OsonDecoder.__new__(OsonDecoder)
            return decoder.decode(data)

    cdef object read_lob_with_length(self, BaseThinConnImpl conn_impl,
                                     DbType dbtype, object lob):
        """
        Read a LOB locator from the buffer and return a LOB object containing
        it.
        """
        cdef:
            uint32_t chunk_size, num_bytes
            BaseThinLobImpl lob_impl
            uint64_t size
            bytes locator
            type cls
        self.read_ub4(&num_bytes)
        if num_bytes > 0:
            if dbtype._ora_type_num == ORA_TYPE_NUM_BFILE:
                size = chunk_size = 0
            else:
                self.read_ub8(&size)
                self.read_ub4(&chunk_size)
            locator = self.read_bytes()
            if lob is None:
                lob_impl = conn_impl._create_lob_impl(dbtype, locator)
                cls = PY_TYPE_ASYNC_LOB \
                        if conn_impl._protocol._transport._is_async \
                        else PY_TYPE_LOB
                lob = cls._from_impl(lob_impl)
            else:
                lob_impl = lob._impl
                lob_impl._locator = locator
            lob_impl._size = size
            lob_impl._chunk_size = chunk_size
            lob_impl._has_metadata = \
                    dbtype._ora_type_num != ORA_TYPE_NUM_BFILE
            return lob

    cdef const char_type* read_raw_bytes(self, ssize_t num_bytes) except NULL:
        """
        Read the specified number of bytes from the packet and return them.
        """
        self._chunked_bytes_buf.start_chunked_read()
        self._get_raw(num_bytes, in_chunked_read=True)
        return self._chunked_bytes_buf.end_chunked_read()

    cdef int read_rowid(self, Rowid *rowid) except -1:
        """
        Reads a rowid from the buffer and populates the rowid structure.
        """
        self.read_ub4(&rowid.rba)
        self.read_ub2(&rowid.partition_id)
        self.skip_ub1()
        self.read_ub4(&rowid.block_num)
        self.read_ub2(&rowid.slot_num)

    cdef object read_urowid(self):
        """
        Read a universal rowid from the buffer and return the Python object
        representing its value.
        """
        cdef:
            ssize_t output_len, input_len, remainder, pos
            int input_offset = 1, output_offset = 0
            const char_type *input_ptr
            bytearray output_value
            uint32_t num_bytes
            uint8_t length
            Rowid rowid

        # get data (first buffer contains the length, which can be ignored)
        self.read_raw_bytes_and_length(&input_ptr, &input_len)
        if input_ptr == NULL:
            return None
        self.read_raw_bytes_and_length(&input_ptr, &input_len)

        # handle physical rowid
        if input_ptr[0] == 1:
            rowid.rba = decode_uint32be(&input_ptr[1])
            rowid.partition_id = decode_uint16be(&input_ptr[5])
            rowid.block_num = decode_uint32be(&input_ptr[7])
            rowid.slot_num = decode_uint16be(&input_ptr[11])
            return _encode_rowid(&rowid)

        # handle logical rowid
        output_len = (input_len // 3) * 4
        remainder = input_len % 3
        if remainder == 1:
            output_len += 1
        elif remainder == 2:
            output_len += 3
        output_value = bytearray(output_len)
        input_len -= 1
        output_value[0] = 42            # '*'
        output_offset += 1
        while input_len > 0:

            # produce first byte of quadruple
            pos = input_ptr[input_offset] >> 2
            output_value[output_offset] = TNS_BASE64_ALPHABET_ARRAY[pos]
            output_offset += 1

            # produce second byte of quadruple, but if only one byte is left,
            # produce that one byte and exit
            pos = (input_ptr[input_offset] & 0x3) << 4
            if input_len == 1:
                output_value[output_offset] = TNS_BASE64_ALPHABET_ARRAY[pos]
                break
            input_offset += 1
            pos |= ((input_ptr[input_offset] & 0xf0) >> 4)
            output_value[output_offset] = TNS_BASE64_ALPHABET_ARRAY[pos]
            output_offset += 1

            # produce third byte of quadruple, but if only two bytes are left,
            # produce that one byte and exit
            pos = (input_ptr[input_offset] & 0xf) << 2
            if input_len == 2:
                output_value[output_offset] = TNS_BASE64_ALPHABET_ARRAY[pos]
                break
            input_offset += 1
            pos |= ((input_ptr[input_offset] & 0xc0) >> 6)
            output_value[output_offset] = TNS_BASE64_ALPHABET_ARRAY[pos]
            output_offset += 1

            # produce final byte of quadruple
            pos = input_ptr[input_offset] & 0x3f
            output_value[output_offset] = TNS_BASE64_ALPHABET_ARRAY[pos]
            output_offset += 1
            input_offset += 1
            input_len -= 3

        return bytes(output_value).decode()

    cdef object read_vector(self):
        """
        Read a VECTOR value from the buffer and return the converted value.
        VECTOR is sent as a LOB value with all of the data prefetched. Since
        the LOB locator is not required it is simply discarded.
        it.
        """
        cdef:
            VectorDecoder decoder
            uint32_t num_bytes
            bytes data
        self.read_ub4(&num_bytes)
        if num_bytes > 0:
            self.skip_ub8()             # size (unused)
            self.skip_ub4()             # chunk size (unused)
            data = self.read_bytes()
            self.read_bytes()           # LOB locator (unused)
            if data:
                decoder = VectorDecoder.__new__(VectorDecoder)
                return decoder.decode(data)

    cdef object read_xmltype(self, BaseThinConnImpl conn_impl):
        """
        Reads an XMLType value from the buffer and returns the string value.
        The XMLType object is a special DbObjectType and is handled separately
        since the structure is a bit different.
        """
        cdef:
            DbObjectPickleBuffer buf
            uint32_t num_bytes
        self.read_ub4(&num_bytes)
        if num_bytes > 0:                   # type OID
            self.read_bytes()
        self.read_ub4(&num_bytes)
        if num_bytes > 0:                   # OID
            self.read_bytes()
        self.read_ub4(&num_bytes)
        if num_bytes > 0:                   # snapshot
            self.read_bytes()
        self.skip_ub2()                     # version
        self.read_ub4(&num_bytes)           # length of data
        self.skip_ub2()                     # flags
        if num_bytes > 0:
            buf = DbObjectPickleBuffer.__new__(DbObjectPickleBuffer)
            buf._populate_from_bytes(self.read_bytes())
            return buf.read_xmltype(conn_impl)

    cdef int check_control_packet(self) except -1:
        """
        Checks for a control packet or final close packet from the server.
        """
        cdef:
            bint notify_waiter
            Packet packet
        packet = self._transport.read_packet()
        self._process_packet(packet, &notify_waiter, False)
        if notify_waiter:
            self._start_packet()

    cdef bint has_response(self):
        """
        Returns a boolean indicating if the list of saved packets contains all
        of the packets for a response from the database. This method can only
        be called if support for the end of response bit is present.
        """
        cdef:
            Packet packet
            ssize_t i, max_pos
        for i in range(self._next_packet_pos, len(self._saved_packets)):
            packet = <Packet> self._saved_packets[i]
            if packet.has_end_of_response():
                return True
        return False

    cdef int reset_packets(self) except -1:
        """
        Resets the list of saved packets and the saved position (called when a
        request has been sent to the database and a response is expected).
        """
        self._saved_packets = []
        self._next_packet_pos = 0
        self._saved_packet_pos = 0
        self._saved_pos = 0

    cdef int restore_point(self) except -1:
        """
        Restores the position in the packets to the last saved point. This is
        needed by asyncio where an ansychronous wait for more packets is
        required so the processing of the response must be restarted at a known
        position.
        """
        if self._saved_packet_pos != self._next_packet_pos - 1:
            self._current_packet = self._saved_packets[self._saved_packet_pos]
            self._populate_from_bytes(self._current_packet.buf)
            self._next_packet_pos = self._saved_packet_pos + 1
        self._pos = self._saved_pos

    cdef int save_point(self) except -1:
        """
        Saves the current position in the packets. This is needed by asyncio
        where an asynchronous wait for more packets is required so the
        processing of the response must be restarted at a known position.
        """
        self._saved_packet_pos = self._next_packet_pos - 1
        self._saved_pos = self._pos

    cdef int skip_raw_bytes_chunked(self) except -1:
        """
        Skip a number of bytes that may or may not be chunked in the buffer.
        The first byte gives the length. If the length is
        TNS_LONG_LENGTH_INDICATOR, however, chunks are read and discarded.
        """
        cdef:
            uint32_t temp_num_bytes
            uint8_t length
        self.read_ub1(&length)
        if length != TNS_LONG_LENGTH_INDICATOR:
            self.skip_raw_bytes(length)
        else:
            while True:
                self.read_ub4(&temp_num_bytes)
                if temp_num_bytes == 0:
                    break
                self.skip_raw_bytes(temp_num_bytes)

    async def wait_for_packets_async(self):
        """
        Wait for packets to arrive in response to the request that was sent to
        the database (using asyncio).
        """
        if self._next_packet_pos >= len(self._saved_packets):
            self._waiter = self._loop.create_future()
            await self._waiter
        self._start_packet()

    cdef int wait_for_packets_sync(self, bint check_marker=False) except -1:
        """
        Wait for packets to arrive in response to the request that was sent to
        the database (synchronously). If no packets are available and we are
        using asyncio, raise an exception so that processing can be restarted
        once packets have arrived. If the check_marker flag is set and a marker
        is detected, throw an exception so that the protocol can process it
        accordingly. This is required because the server can send a marker
        packet in the middle of the data packets that form the response to the
        client's request.
        """
        cdef:
            bint notify_waiter
            Packet packet
        if self._next_packet_pos >= len(self._saved_packets):
            if self._transport._is_async:
                raise OutOfPackets()
            while True:
                packet = self._transport.read_packet()
                self._process_packet(packet, &notify_waiter, True)
                if notify_waiter:
                    break
        self._start_packet()
        if check_marker \
                and self._current_packet.packet_type == TNS_PACKET_TYPE_MARKER:
            raise MarkerDetected()

    async def wait_for_response_async(self):
        """
        Wait for packets to arrive in response to the request that was sent to
        the database (using asyncio). This method will not return until the
        complete response has been received. This requires the "end of
        response" capability available in Oracle Database 23ai and higher. This
        method also assumes that the current list of saved packets does not
        contain a full response.
        """
        try:
            self._check_request_boundary = True
            self._waiter = self._loop.create_future()
            await self._waiter
        finally:
            self._check_request_boundary = False


@cython.final
cdef class WriteBuffer(Buffer):

    cdef:
        uint8_t _packet_type
        uint8_t _packet_flags
        uint16_t _data_flags
        Capabilities _caps
        Transport _transport
        uint8_t _seq_num
        bint _packet_sent

    def __cinit__(self, Transport transport, Capabilities caps):
        self._transport = transport
        self._caps = caps
        self._size_for_sdu()

    cdef int _send_packet(self, bint final_packet) except -1:
        """
        Write the packet header and then send the packet. Once sent, reset the
        pointers back to an empty packet.
        """
        cdef ssize_t size = self._pos
        self._pos = 0
        if self._caps.protocol_version >= TNS_VERSION_MIN_LARGE_SDU:
            self.write_uint32be(size)
        else:
            self.write_uint16be(size)
            self.write_uint16be(0)
        self.write_uint8(self._packet_type)
        self.write_uint8(self._packet_flags)
        self.write_uint16be(0)
        if self._packet_type == TNS_PACKET_TYPE_DATA:
            self.write_uint16be(self._data_flags)
        self._pos = size
        self._transport.write_packet(self)
        self._packet_sent = True
        self._pos = PACKET_HEADER_SIZE
        if not final_packet and self._packet_type == TNS_PACKET_TYPE_DATA:
            self._pos += sizeof(uint16_t)   # allow space for data flags

    cdef int _size_for_sdu(self) except -1:
        """
        Resizes the buffer based on the SDU size of the capabilities.
        """
        self._initialize(self._caps.sdu)

    cdef int _write_more_data(self, ssize_t num_bytes_available,
                              ssize_t num_bytes_wanted) except -1:
        """
        Called when the amount of buffer available is less than the amount of
        data requested. This sends the packet to the server and then resets the
        buffer for further writing.
        """
        self._send_packet(final_packet=False)

    cdef int end_request(self) except -1:
        """
        Indicates that the request from the client is completing and will send
        any packet remaining, if necessary.
        """
        if self._pos > PACKET_HEADER_SIZE:
            self._send_packet(final_packet=True)

    cdef inline ssize_t max_payload_bytes(self):
        """
        Return the maximum number of bytes that can be sent in a packet. This
        is the maximum size of the entire packet, less the bytes in the header
        and 2 bytes for the data flags.
        """
        return self._max_size - PACKET_HEADER_SIZE - 2

    cdef void start_request(self, uint8_t packet_type, uint8_t packet_flags=0,
                            uint16_t data_flags=0):
        """
        Indicates that a request from the client is starting. The packet type
        is retained just in case a request spans multiple packets. The packet
        header (8 bytes in length) is written when a packet is actually being
        sent and so is skipped at this point.
        """
        self._packet_sent = False
        self._packet_type = packet_type
        self._packet_flags = packet_flags
        self._pos = PACKET_HEADER_SIZE
        if packet_type == TNS_PACKET_TYPE_DATA:
            self._data_flags = data_flags
            self._pos += sizeof(uint16_t)

    cdef object write_dbobject(self, ThinDbObjectImpl obj_impl):
        """
        Writes a database object to the buffer.
        """
        cdef:
            ThinDbObjectTypeImpl typ_impl = obj_impl.type
            uint32_t num_bytes
            bytes packed_data
        self.write_ub4(len(obj_impl.toid))
        self.write_bytes_with_length(obj_impl.toid)
        if obj_impl.oid is None:
            self.write_ub4(0)
        else:
            self.write_ub4(len(obj_impl.oid))
            self.write_bytes_with_length(obj_impl.oid)
        self.write_ub4(0)                   # snapshot
        self.write_ub4(0)                   # version
        packed_data = obj_impl._get_packed_data()
        self.write_ub4(len(packed_data))
        self.write_ub4(obj_impl.flags)      # flags
        self.write_bytes_with_length(packed_data)

    cdef int write_extension_values(self, str txt_value, bytes bytes_value,
                                    uint16_t keyword) except -1:
        """
        Writes extension's text value, binary value and keyword entry to the
        buffer.
        """
        cdef bytes txt_value_bytes
        if txt_value is None:
            self.write_uint8(0)
        else:
            txt_value_bytes = txt_value.encode()
            self.write_ub4(len(txt_value_bytes))
            self.write_bytes_with_length(txt_value_bytes)
        if bytes_value is None:
            self.write_uint8(0)
        else:
            self.write_ub4(len(bytes_value))
            self.write_bytes_with_length(bytes_value)
        self.write_ub2(keyword)

    cdef int write_lob_with_length(self, BaseThinLobImpl lob_impl) except -1:
        """
        Writes a LOB locator to the buffer.
        """
        self.write_ub4(len(lob_impl._locator))
        self.write_bytes_with_length(lob_impl._locator)

    cdef int write_qlocator(self, uint64_t data_length,
                            bint write_length=True) except -1:
        """
        Writes a QLocator. QLocators are always 40 bytes in length.
        """
        self.write_ub4(40)                  # QLocator length
        if write_length:
            self.write_uint8(40)            # chunk length
        self.write_uint16be(38)             # QLocator length less 2 bytes
        self.write_uint16be(TNS_LOB_QLOCATOR_VERSION)
        self.write_uint8(TNS_LOB_LOC_FLAGS_VALUE_BASED | \
                         TNS_LOB_LOC_FLAGS_BLOB | \
                         TNS_LOB_LOC_FLAGS_ABSTRACT)
        self.write_uint8(TNS_LOB_LOC_FLAGS_INIT)
        self.write_uint16be(0)              # additional flags
        self.write_uint16be(1)              # byt1
        self.write_uint64be(data_length)
        self.write_uint16be(0)              # unused
        self.write_uint16be(0)              # csid
        self.write_uint16be(0)              # unused
        self.write_uint64be(0)              # unused
        self.write_uint64be(0)              # unused

    cdef object write_oson(self, value, ssize_t max_fname_size,
                           bint write_length=True):
        """
        Encodes the given value to OSON and then writes that to the buffer.
        it.
        """
        cdef OsonEncoder encoder = OsonEncoder.__new__(OsonEncoder)
        encoder.encode(value, max_fname_size)
        self.write_qlocator(encoder._pos, write_length)
        self._write_raw_bytes_and_length(encoder._data, encoder._pos)

    cdef int write_seq_num(self) except -1:
        self._seq_num += 1
        if self._seq_num == 0:
            self._seq_num = 1
        self.write_uint8(self._seq_num)

    cdef object write_vector(self, value):
        """
        Encodes the given value to VECTOR and then writes that to the buffer.
        """
        cdef VectorEncoder encoder = VectorEncoder.__new__(VectorEncoder)
        encoder.encode(value)
        self.write_qlocator(encoder._pos)
        self._write_raw_bytes_and_length(encoder._data, encoder._pos)



# -----------------------------------------------------------------------------
# Copyright (c) 2020, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
2300 - Module for testing object variables
"""

import datetime
import decimal
import unittest

import oracledb
import test_env


class TestCase(test_env.BaseTestCase):
    def __test_data(
        self, expected_int_value, expected_obj_value, expected_array_value
    ):
        int_value, object_value, array_value = self.cursor.fetchone()
        if object_value is not None:
            self.assertIsInstance(object_value.INTVALUE, int)
            self.assertIsInstance(object_value.SMALLINTVALUE, int)
            self.assertIsInstance(object_value.FLOATVALUE, float)
        if object_value is not None:
            object_value = self.get_db_object_as_plain_object(object_value)
        if array_value is not None:
            array_value = array_value.aslist()
        self.assertEqual(int_value, expected_int_value)
        self.assertEqual(object_value, expected_obj_value)
        self.assertEqual(array_value, expected_array_value)

    def test_2300(self):
        "2300 - test binding a null value (IN)"
        var = self.cursor.var(oracledb.DB_TYPE_OBJECT, typename="UDT_OBJECT")
        result = self.cursor.callfunc(
            "pkg_TestBindObject.GetStringRep", str, [var]
        )
        self.assertEqual(result, "null")

    def test_2301(self):
        "2301 - test binding an object (IN)"
        type_obj = self.conn.gettype("UDT_OBJECT")
        obj = type_obj.newobject()
        obj.NUMBERVALUE = 13
        obj.STRINGVALUE = "Test String"
        result = self.cursor.callfunc(
            "pkg_TestBindObject.GetStringRep", str, [obj]
        )
        exp = "udt_Object(13, 'Test String', null, null, null, null, null)"
        self.assertEqual(result, exp)
        obj.NUMBERVALUE = None
        obj.STRINGVALUE = "Test With Dates"
        obj.DATEVALUE = datetime.datetime(2016, 2, 10)
        obj.TIMESTAMPVALUE = datetime.datetime(2016, 2, 10, 14, 13, 50)
        result = self.cursor.callfunc(
            "pkg_TestBindObject.GetStringRep", str, [obj]
        )
        expected_value = (
            "udt_Object(null, 'Test With Dates', null, "
            "to_date('2016-02-10', 'YYYY-MM-DD'), "
            "to_timestamp('2016-02-10 14:13:50', "
            "'YYYY-MM-DD HH24:MI:SS'), "
            "null, null)"
        )
        self.assertEqual(result, expected_value)
        obj.DATEVALUE = None
        obj.TIMESTAMPVALUE = None
        sub_type_obj = self.conn.gettype("UDT_SUBOBJECT")
        sub_obj = sub_type_obj.newobject()
        sub_obj.SUBNUMBERVALUE = decimal.Decimal("18.25")
        sub_obj.SUBSTRINGVALUE = "Sub String"
        obj.SUBOBJECTVALUE = sub_obj
        result = self.cursor.callfunc(
            "pkg_TestBindObject.GetStringRep", str, [obj]
        )
        expected_value = (
            "udt_Object(null, 'Test With Dates', null, null, "
            "null, udt_SubObject(18.25, 'Sub String'), null)"
        )
        self.assertEqual(result, expected_value)

    def test_2302(self):
        "2302 - test copying an object"
        type_obj = self.conn.gettype("UDT_OBJECT")
        obj = type_obj()
        obj.NUMBERVALUE = 5124
        obj.STRINGVALUE = "A test string"
        obj.DATEVALUE = datetime.datetime(2016, 2, 24)
        obj.TIMESTAMPVALUE = datetime.datetime(2016, 2, 24, 13, 39, 10)
        copied_obj = obj.copy()
        self.assertEqual(obj.NUMBERVALUE, copied_obj.NUMBERVALUE)
        self.assertEqual(obj.STRINGVALUE, copied_obj.STRINGVALUE)
        self.assertEqual(obj.DATEVALUE, copied_obj.DATEVALUE)
        self.assertEqual(obj.TIMESTAMPVALUE, copied_obj.TIMESTAMPVALUE)

    def test_2303(self):
        "2303 - test getting an empty collection as a list"
        type_obj = self.conn.gettype("UDT_ARRAY")
        obj = type_obj.newobject()
        self.assertEqual(obj.aslist(), [])

    def test_2304(self):
        "2304 - test fetching objects"
        self.cursor.execute(
            """
            select IntCol, ObjectCol, ArrayCol
            from TestObjects
            order by IntCol
            """
        )
        expected_value = [
            ("INTCOL", oracledb.DB_TYPE_NUMBER, 10, None, 9, 0, False),
            (
                "OBJECTCOL",
                oracledb.DB_TYPE_OBJECT,
                None,
                None,
                None,
                None,
                True,
            ),
            (
                "ARRAYCOL",
                oracledb.DB_TYPE_OBJECT,
                None,
                None,
                None,
                None,
                True,
            ),
        ]
        self.assertEqual(self.cursor.description, expected_value)
        expected_value = (
            1,
            "First row",
            "First     ",
            "N First Row",
            "N First   ",
            b"Raw Data 1",
            2,
            5,
            12.125,
            0.5,
            12.5,
            25.25,
            50.125,
            datetime.datetime(2007, 3, 6, 0, 0, 0),
            datetime.datetime(2008, 9, 12, 16, 40),
            datetime.datetime(2009, 10, 13, 17, 50),
            oracledb.Timestamp(2010, 11, 14, 18, 55),
            "Short CLOB value",
            "Short NCLOB Value",
            b"Short BLOB value",
            (11, "Sub object 1"),
            [(5, "first element"), (6, "second element")],
        )
        self.__test_data(1, expected_value, [5, 10, None, 20])
        self.__test_data(2, None, [3, None, 9, 12, 15])
        expected_value = (
            3,
            "Third row",
            "Third     ",
            "N Third Row",
            "N Third   ",
            b"Raw Data 3",
            4,
            10,
            6.5,
            0.75,
            43.25,
            86.5,
            192.125,
            datetime.datetime(2007, 6, 21, 0, 0, 0),
            datetime.datetime(2007, 12, 13, 7, 30, 45),
            datetime.datetime(2017, 6, 21, 23, 18, 45),
            oracledb.Timestamp(2017, 7, 21, 8, 27, 13),
            "Another short CLOB value",
            "Another short NCLOB Value",
            b"Yet another short BLOB value",
            (13, "Sub object 3"),
            [
                (10, "element #1"),
                (20, "element #2"),
                (30, "element #3"),
                (40, "element #4"),
            ],
        )
        self.__test_data(3, expected_value, None)

    def test_2305(self):
        "2305 - test getting object type"
        type_obj = self.conn.gettype("UDT_OBJECT")
        self.assertFalse(type_obj.iscollection)
        self.assertEqual(type_obj.schema, self.conn.username.upper())
        self.assertEqual(type_obj.name, "UDT_OBJECT")
        sub_object_value_type = self.conn.gettype("UDT_SUBOBJECT")
        sub_object_array_type = self.conn.gettype("UDT_OBJECTARRAY")
        expected_metadata = [
            ("NUMBERVALUE", oracledb.DB_TYPE_NUMBER, 0, -127, None),
            ("STRINGVALUE", oracledb.DB_TYPE_VARCHAR, None, None, 60),
            ("FIXEDCHARVALUE", oracledb.DB_TYPE_CHAR, None, None, 10),
            ("NSTRINGVALUE", oracledb.DB_TYPE_NVARCHAR, None, None, 120),
            ("NFIXEDCHARVALUE", oracledb.DB_TYPE_NCHAR, None, None, 20),
            ("RAWVALUE", oracledb.DB_TYPE_RAW, None, None, 16),
            ("INTVALUE", oracledb.DB_TYPE_NUMBER, 38, 0, None),
            ("SMALLINTVALUE", oracledb.DB_TYPE_NUMBER, 38, 0, None),
            ("REALVALUE", oracledb.DB_TYPE_NUMBER, 63, -127, None),
            ("DOUBLEPRECISIONVALUE", oracledb.DB_TYPE_NUMBER, 126, -127, None),
            ("FLOATVALUE", oracledb.DB_TYPE_NUMBER, 126, -127, None),
            (
                "BINARYFLOATVALUE",
                oracledb.DB_TYPE_BINARY_FLOAT,
                None,
                None,
                None,
            ),
            (
                "BINARYDOUBLEVALUE",
                oracledb.DB_TYPE_BINARY_DOUBLE,
                None,
                None,
                None,
            ),
            ("DATEVALUE", oracledb.DB_TYPE_DATE, None, None, None),
            ("TIMESTAMPVALUE", oracledb.DB_TYPE_TIMESTAMP, None, None, None),
            (
                "TIMESTAMPTZVALUE",
                oracledb.DB_TYPE_TIMESTAMP_TZ,
                None,
                None,
                None,
            ),
            (
                "TIMESTAMPLTZVALUE",
                oracledb.DB_TYPE_TIMESTAMP_LTZ,
                None,
                None,
                None,
            ),
            ("CLOBVALUE", oracledb.DB_TYPE_CLOB, None, None, None),
            ("NCLOBVALUE", oracledb.DB_TYPE_NCLOB, None, None, None),
            ("BLOBVALUE", oracledb.DB_TYPE_BLOB, None, None, None),
            ("SUBOBJECTVALUE", sub_object_value_type, None, None, None),
            ("SUBOBJECTARRAY", sub_object_array_type, None, None, None),
        ]
        actual_metadata = [
            (attr.name, attr.type, attr.precision, attr.scale, attr.max_size)
            for attr in type_obj.attributes
        ]
        self.assertEqual(actual_metadata, expected_metadata)
        self.assertEqual(sub_object_array_type.iscollection, True)
        self.assertEqual(sub_object_array_type.attributes, [])

    def test_2306(self):
        "2306 - test object type data"
        self.cursor.execute(
            """
            select ObjectCol
            from TestObjects
            where ObjectCol is not null
              and rownum <= 1
            """
        )
        (obj,) = self.cursor.fetchone()
        self.assertEqual(obj.type.schema, self.conn.username.upper())
        self.assertEqual(obj.type.name, "UDT_OBJECT")
        self.assertEqual(obj.type.attributes[0].name, "NUMBERVALUE")

    def test_2307(self):
        "2307 - test inserting and then querying object with all data types"
        self.cursor.execute("delete from TestClobs")
        self.cursor.execute("delete from TestNClobs")
        self.cursor.execute("delete from TestBlobs")
        self.cursor.execute("delete from TestObjects where IntCol > 3")
        self.cursor.execute(
            """
            insert into TestClobs (IntCol, ClobCol)
            values (1, 'A short CLOB')
            """
        )
        self.cursor.execute(
            """
            insert into TestNClobs (IntCol, NClobCol)
            values (1, 'A short NCLOB')
            """
        )
        self.cursor.execute(
            """
            insert into TestBlobs (IntCol, BlobCol)
            values (1, utl_raw.cast_to_raw('A short BLOB'))
            """
        )
        self.conn.commit()
        self.cursor.execute("select CLOBCol from TestClobs")
        (clob,) = self.cursor.fetchone()
        self.cursor.execute("select NCLOBCol from TestNClobs")
        (nclob,) = self.cursor.fetchone()
        self.cursor.execute("select BLOBCol from TestBlobs")
        (blob,) = self.cursor.fetchone()
        type_obj = self.conn.gettype("UDT_OBJECT")
        obj = type_obj.newobject()
        obj.NUMBERVALUE = 5
        obj.STRINGVALUE = "A string"
        obj.FIXEDCHARVALUE = "Fixed str"
        obj.NSTRINGVALUE = "A NCHAR string"
        obj.NFIXEDCHARVALUE = "Fixed N"
        obj.RAWVALUE = b"Raw Value"
        obj.INTVALUE = 27
        obj.SMALLINTVALUE = 13
        obj.REALVALUE = 184.875
        obj.DOUBLEPRECISIONVALUE = 1.375
        obj.FLOATVALUE = 23.0
        obj.DATEVALUE = datetime.date(2017, 5, 9)
        obj.TIMESTAMPVALUE = datetime.datetime(2017, 5, 9, 9, 41, 13)
        obj.TIMESTAMPTZVALUE = datetime.datetime(1986, 8, 2, 15, 27, 38)
        obj.TIMESTAMPLTZVALUE = datetime.datetime(1999, 11, 12, 23, 5, 2)
        obj.BINARYFLOATVALUE = 14.25
        obj.BINARYDOUBLEVALUE = 29.1625
        obj.CLOBVALUE = clob
        obj.NCLOBVALUE = nclob
        obj.BLOBVALUE = blob
        sub_type_obj = self.conn.gettype("UDT_SUBOBJECT")
        sub_obj = sub_type_obj.newobject()
        sub_obj.SUBNUMBERVALUE = 23
        sub_obj.SUBSTRINGVALUE = "Substring value"
        obj.SUBOBJECTVALUE = sub_obj
        self.cursor.execute(
            """
            insert into TestObjects (IntCol, ObjectCol)
            values (4, :obj)
            """,
            obj=obj,
        )
        self.cursor.execute(
            """
            select IntCol, ObjectCol, ArrayCol
            from TestObjects
            where IntCol = 4
            """
        )
        expected_value = (
            5,
            "A string",
            "Fixed str ",
            "A NCHAR string",
            "Fixed N   ",
            b"Raw Value",
            27,
            13,
            184.875,
            1.375,
            23.0,
            14.25,
            29.1625,
            datetime.datetime(2017, 5, 9, 0, 0, 0),
            datetime.datetime(2017, 5, 9, 9, 41, 13),
            datetime.datetime(1986, 8, 2, 15, 27, 38),
            oracledb.Timestamp(1999, 11, 12, 23, 5, 2),
            "A short CLOB",
            "A short NCLOB",
            b"A short BLOB",
            (23, "Substring value"),
            None,
        )
        self.__test_data(4, expected_value, None)
        obj.CLOBVALUE = "A short CLOB (modified)"
        obj.NCLOBVALUE = "A short NCLOB (modified)"
        obj.BLOBVALUE = "A short BLOB (modified)"
        self.cursor.execute(
            """
            insert into TestObjects (IntCol, ObjectCol)
            values (5, :obj)
            """,
            obj=obj,
        )
        self.cursor.execute(
            """
            select IntCol, ObjectCol, ArrayCol
            from TestObjects
            where IntCol = 5
            """
        )
        expected_value = (
            5,
            "A string",
            "Fixed str ",
            "A NCHAR string",
            "Fixed N   ",
            b"Raw Value",
            27,
            13,
            184.875,
            1.375,
            23.0,
            14.25,
            29.1625,
            datetime.datetime(2017, 5, 9, 0, 0, 0),
            datetime.datetime(2017, 5, 9, 9, 41, 13),
            datetime.datetime(1986, 8, 2, 15, 27, 38),
            oracledb.Timestamp(1999, 11, 12, 23, 5, 2),
            "A short CLOB (modified)",
            "A short NCLOB (modified)",
            b"A short BLOB (modified)",
            (23, "Substring value"),
            None,
        )
        self.__test_data(5, expected_value, None)
        self.conn.rollback()

    def test_2308(self):
        "2308 - test trying to find an object type that does not exist"
        self.assertRaises(TypeError, self.conn.gettype, 2)
        with self.assertRaisesFullCode("DPY-2035"):
            self.conn.gettype("A TYPE THAT DOES NOT EXIST")

    def test_2309(self):
        "2309 - test appending an object of the wrong type to a collection"
        collection_obj_type = self.conn.gettype("UDT_OBJECTARRAY")
        collection_obj = collection_obj_type.newobject()
        array_obj_type = self.conn.gettype("UDT_ARRAY")
        array_obj = array_obj_type.newobject()
        with self.assertRaisesFullCode("DPY-2008"):
            collection_obj.append(array_obj)

    def test_2310(self):
        "2310 - test that referencing a sub object affects the parent object"
        obj_type = self.conn.gettype("UDT_OBJECT")
        sub_obj_type = self.conn.gettype("UDT_SUBOBJECT")
        obj = obj_type.newobject()
        obj.SUBOBJECTVALUE = sub_obj_type.newobject()
        obj.SUBOBJECTVALUE.SUBNUMBERVALUE = 5
        obj.SUBOBJECTVALUE.SUBSTRINGVALUE = "Substring"
        self.assertEqual(obj.SUBOBJECTVALUE.SUBNUMBERVALUE, 5)
        self.assertEqual(obj.SUBOBJECTVALUE.SUBSTRINGVALUE, "Substring")

    def test_2311(self):
        "2311 - test accessing sub object after parent object destroyed"
        obj_type = self.conn.gettype("UDT_OBJECT")
        sub_obj_type = self.conn.gettype("UDT_SUBOBJECT")
        array_type = self.conn.gettype("UDT_OBJECTARRAY")
        sub_obj1 = sub_obj_type.newobject()
        sub_obj1.SUBNUMBERVALUE = 2
        sub_obj1.SUBSTRINGVALUE = "AB"
        sub_obj2 = sub_obj_type.newobject()
        sub_obj2.SUBNUMBERVALUE = 3
        sub_obj2.SUBSTRINGVALUE = "CDE"
        obj = obj_type.newobject()
        obj.SUBOBJECTARRAY = array_type.newobject([sub_obj1, sub_obj2])
        sub_obj_array = obj.SUBOBJECTARRAY
        del obj
        self.assertEqual(
            self.get_db_object_as_plain_object(sub_obj_array),
            [(2, "AB"), (3, "CDE")],
        )

    def test_2312(self):
        "2312 - test assigning an object of wrong type to an object attribute"
        obj_type = self.conn.gettype("UDT_OBJECT")
        obj = obj_type.newobject()
        wrong_obj_type = self.conn.gettype("UDT_OBJECTARRAY")
        wrong_obj = wrong_obj_type.newobject()
        with self.assertRaisesFullCode("DPY-2008"):
            setattr(obj, "SUBOBJECTVALUE", wrong_obj)

    def test_2313(self):
        "2313 - test setting value of object variable to wrong object type"
        wrong_obj_type = self.conn.gettype("UDT_OBJECTARRAY")
        wrong_obj = wrong_obj_type.newobject()
        var = self.cursor.var(oracledb.DB_TYPE_OBJECT, typename="UDT_OBJECT")
        with self.assertRaisesFullCode("DPY-2008"):
            var.setvalue(0, wrong_obj)

    def test_2314(self):
        "2314 - test object string format"
        obj_type = self.conn.gettype("UDT_OBJECT")
        user = test_env.get_main_user()
        self.assertEqual(
            str(obj_type), f"<oracledb.DbObjectType {user.upper()}.UDT_OBJECT>"
        )
        self.assertEqual(
            str(obj_type.attributes[0]), "<oracledb.DbObjectAttr NUMBERVALUE>"
        )

    def test_2315(self):
        "2315 - test Trim number of elements from collection"
        sub_obj_type = self.conn.gettype("UDT_SUBOBJECT")
        array_type = self.conn.gettype("UDT_OBJECTARRAY")
        data = [(1, "AB"), (2, "CDE"), (3, "FGH"), (4, "IJK"), (5, "LMN")]
        array_obj = array_type()
        for num_val, str_val in data:
            subObj = sub_obj_type()
            subObj.SUBNUMBERVALUE = num_val
            subObj.SUBSTRINGVALUE = str_val
            array_obj.append(subObj)
        self.assertEqual(self.get_db_object_as_plain_object(array_obj), data)
        array_obj.trim(2)
        self.assertEqual(
            self.get_db_object_as_plain_object(array_obj), data[:3]
        )
        array_obj.trim(1)
        self.assertEqual(
            self.get_db_object_as_plain_object(array_obj), data[:2]
        )
        array_obj.trim(0)
        self.assertEqual(
            self.get_db_object_as_plain_object(array_obj), data[:2]
        )
        array_obj.trim(2)
        self.assertEqual(self.get_db_object_as_plain_object(array_obj), [])

    def test_2316(self):
        "2316 - test the metadata of a SQL type"
        user = test_env.get_main_user()
        typ = self.conn.gettype("UDT_OBJECTARRAY")
        self.assertEqual(typ.schema, user.upper())
        self.assertEqual(typ.name, "UDT_OBJECTARRAY")
        self.assertIsNone(typ.package_name)
        self.assertEqual(typ.element_type.schema, user.upper())
        self.assertEqual(typ.element_type.name, "UDT_SUBOBJECT")
        self.assertIsNone(typ.element_type.package_name)
        self.assertEqual(typ.attributes, [])
        self.assertTrue(typ.iscollection)

    def test_2317(self):
        "2317 - test the metadata of a PL/SQL type"
        user = test_env.get_main_user()
        typ = self.conn.gettype("PKG_TESTSTRINGARRAYS.UDT_STRINGLIST")
        self.assertEqual(typ.schema, user.upper())
        self.assertEqual(typ.name, "UDT_STRINGLIST")
        self.assertEqual(typ.package_name, "PKG_TESTSTRINGARRAYS")
        self.assertEqual(typ.element_type, oracledb.DB_TYPE_VARCHAR)
        self.assertEqual(typ.attributes, [])
        self.assertTrue(typ.iscollection)

    def test_2318(self):
        "2318 - test creating an object variable without a type name"
        with self.assertRaisesFullCode("DPY-2037"):
            self.cursor.var(oracledb.DB_TYPE_OBJECT)

    def test_2319(self):
        "2319 - test getting an empty collection as a dictionary"
        type_obj = self.conn.gettype("UDT_ARRAY")
        obj = type_obj.newobject()
        self.assertEqual(obj.asdict(), {})

    def test_2320(self):
        "2320 - test if an element exists in a collection"
        array_type = self.conn.gettype("UDT_ARRAY")
        array_obj = array_type.newobject()
        self.assertFalse(array_obj.exists(0))
        array_obj.append(40)
        self.assertTrue(array_obj.exists(0))
        self.assertFalse(array_obj.exists(1))

    def test_2321(self):
        "2321 - test first and last methods"
        array_type = self.conn.gettype("UDT_ARRAY")
        array_obj = array_type.newobject()
        self.assertIsNone(array_obj.first())
        self.assertIsNone(array_obj.last())
        for i in range(7):
            array_obj.append(i)
        self.assertEqual(array_obj.first(), 0)
        self.assertEqual(array_obj.last(), 6)

    def test_2322(self):
        "2322 - test getting the size of a collections"
        array_type = self.conn.gettype("UDT_ARRAY")
        array_obj = array_type.newobject()
        self.assertEqual(array_obj.size(), 0)
        for i in range(5):
            array_obj.append(i)
        self.assertEqual(array_obj.size(), 5)

    def test_2323(self):
        "2323 - test prev and next methods"
        array_type = self.conn.gettype("UDT_ARRAY")
        array_obj = array_type.newobject()
        self.assertIsNone(array_obj.prev(0))
        self.assertIsNone(array_obj.next(0))
        for i in range(2):
            array_obj.append(i)
        self.assertIsNone(array_obj.prev(0))
        self.assertEqual(array_obj.prev(1), 0)
        self.assertEqual(array_obj.next(0), 1)
        self.assertIsNone(array_obj.next(1))

    def test_2324(self):
        "2324 - test setting and getting elements from a collection"
        array_type = self.conn.gettype("UDT_ARRAY")
        array_obj = array_type.newobject()
        with self.assertRaisesFullCode("DPY-2038"):
            array_obj.getelement(0)
        with self.assertRaisesFullCode("DPY-2039"):
            array_obj.setelement(0, 7)
        array_obj.append(7)
        self.assertEqual(array_obj.getelement(0), 7)
        array_obj.setelement(0, 10)
        self.assertEqual(array_obj.getelement(0), 10)
        with self.assertRaisesFullCode("DPY-2039"):
            array_obj.setelement(3, 4)

    def test_2325(self):
        "2325 - test appending too many elements to a collection"
        array_type = self.conn.gettype("UDT_ARRAY")
        numbers = [i for i in range(11)]
        with self.assertRaisesFullCode("DPY-2039"):
            array_type.newobject(numbers)

        array_obj = array_type.newobject()
        with self.assertRaisesFullCode("DPY-2039"):
            array_obj.extend(numbers)

        array_obj = array_type.newobject()
        for elem in numbers[:10]:
            array_obj.append(elem)
        with self.assertRaisesFullCode("DPY-2039"):
            array_obj.append(numbers[10])

    def test_2326(self):
        "2326 - test appending elements to an unconstrained table"
        data = [1, 3, 6, 10, 15, 21]
        typ = self.conn.gettype("UDT_UNCONSTRAINEDTABLE")
        obj = typ.newobject(data)
        self.cursor.execute("select :1 from dual", [obj])
        (output_obj,) = self.cursor.fetchone()
        self.assertEqual(output_obj.aslist(), data)

    def test_2327(self):
        "2327 - test collection with thousands of entries"
        typ = self.conn.gettype("PKG_TESTNUMBERARRAYS.UDT_NUMBERLIST")
        obj = typ.newobject()
        obj.setelement(1, 1)
        running_total = 1
        for i in range(1, 35000):
            running_total += i + 1
            obj.append(running_total)
        result = self.cursor.callfunc(
            "pkg_TestNumberArrays.TestInArrays", int, (2327, obj)
        )
        self.assertEqual(result, 7146445847327)

    @unittest.skipIf(test_env.get_is_thin(), "thin mode supports xmltype")
    def test_2328(self):
        "2328 - test object with unknown type in one of its attributes"
        typ = self.conn.gettype("UDT_OBJECTWITHXMLTYPE")
        self.assertEqual(typ.attributes[1].type, oracledb.DB_TYPE_UNKNOWN)

    @unittest.skipIf(test_env.get_is_thin(), "thin mode supports xmltype")
    def test_2329(self):
        "2329 - test object with unknown type as the element type"
        typ = self.conn.gettype("UDT_XMLTYPEARRAY")
        self.assertEqual(typ.element_type, oracledb.DB_TYPE_UNKNOWN)

    def test_2330(self):
        "2330 - test DB Object repr()"
        typ = self.conn.gettype("UDT_ARRAY")
        obj = typ.newobject()
        fqn = f"{typ.schema}.{typ.name}"
        expected_str = f"^<oracledb.DbObject {fqn} at 0x.+>$"
        self.assertRegex(repr(obj), expected_str)

        # object of a package
        typ = self.conn.gettype("PKG_TESTSTRINGARRAYS.UDT_STRINGLIST")
        obj = typ.newobject()
        fqn = f"{typ.schema}.{typ.package_name}.{typ.name}"
        expected_str = f"^<oracledb.DbObject {fqn} at 0x.+>$"
        self.assertRegex(repr(obj), expected_str)

    def test_2331(self):
        "2331 - test creating an object with invalid data type"
        type_obj = self.conn.gettype("UDT_ARRAY")
        with self.assertRaisesFullCode("DPY-3013"):
            type_obj.newobject([490, "not a number"])
        with self.assertRaisesFullCode("DPY-3013"):
            type_obj([71, "not a number"])

    def test_2332(self):
        "2332 - test getting an invalid attribute name from an object"
        typ = self.conn.gettype("UDT_OBJECT")
        obj = typ.newobject()
        with self.assertRaises(AttributeError):
            obj.MISSING

    def test_2333(self):
        "2333 - test validating a string attribute"
        typ = self.conn.gettype("UDT_OBJECT")
        obj = typ.newobject()
        for attr_name, max_size in [
            ("STRINGVALUE", 60),
            ("FIXEDCHARVALUE", 10),
            ("NSTRINGVALUE", 120),
            ("NFIXEDCHARVALUE", 20),
            ("RAWVALUE", 16),
        ]:
            with self.subTest(attr_name=attr_name, max_size=max_size):
                value = "A" * max_size
                setattr(obj, attr_name, value)
                value += "X"
                with self.assertRaisesFullCode("DPY-2043"):
                    setattr(obj, attr_name, value)

    def test_2334(self):
        "2334 - test validating a string element value"
        typ = self.conn.gettype("PKG_TESTSTRINGARRAYS.UDT_STRINGLIST")
        obj = typ.newobject()
        obj.append("A" * 100)
        with self.assertRaisesFullCode("DPY-2044"):
            obj.append("A" * 101)
        obj.append("B" * 100)
        with self.assertRaisesFullCode("DPY-2044"):
            obj.setelement(2, "C" * 101)

    def test_2335(self):
        "2335 - test validating a string attribute with null value"
        typ = self.conn.gettype("UDT_OBJECT")
        obj = typ.newobject()
        obj.STRINGVALUE = None

    def test_2336(self):
        "2336 - test initializing (with a sequence) a non collection obj"
        obj_type = self.conn.gettype("UDT_OBJECT")
        with self.assertRaisesFullCode("DPY-2036"):
            obj_type.newobject([1, 2])
        with self.assertRaisesFullCode("DPY-2036"):
            obj_type([3, 4])

    def test_2337(self):
        "2337 - test %ROWTYPE with all types"
        sub_obj_type = self.conn.gettype("UDT_SUBOBJECT")
        sub_arr_type = self.conn.gettype("UDT_OBJECTARRAY")
        expected_metadata = [
            ("NUMBERVALUE", oracledb.DB_TYPE_NUMBER, 0, -127, None),
            ("STRINGVALUE", oracledb.DB_TYPE_VARCHAR, None, None, 60),
            ("FIXEDCHARVALUE", oracledb.DB_TYPE_CHAR, None, None, 10),
            ("NSTRINGVALUE", oracledb.DB_TYPE_NVARCHAR, None, None, 120),
            ("NFIXEDCHARVALUE", oracledb.DB_TYPE_NCHAR, None, None, 20),
            ("RAWVALUE", oracledb.DB_TYPE_RAW, None, None, 16),
            ("INTVALUE", oracledb.DB_TYPE_NUMBER, 38, 0, None),
            ("SMALLINTVALUE", oracledb.DB_TYPE_NUMBER, 38, 0, None),
            ("REALVALUE", oracledb.DB_TYPE_NUMBER, 63, -127, None),
            ("DOUBLEPRECISIONVALUE", oracledb.DB_TYPE_NUMBER, 126, -127, None),
            ("FLOATVALUE", oracledb.DB_TYPE_NUMBER, 126, -127, None),
            (
                "BINARYFLOATVALUE",
                oracledb.DB_TYPE_BINARY_FLOAT,
                None,
                None,
                None,
            ),
            (
                "BINARYDOUBLEVALUE",
                oracledb.DB_TYPE_BINARY_DOUBLE,
                None,
                None,
                None,
            ),
            ("DATEVALUE", oracledb.DB_TYPE_DATE, None, None, None),
            ("TIMESTAMPVALUE", oracledb.DB_TYPE_TIMESTAMP, None, None, None),
            (
                "TIMESTAMPTZVALUE",
                oracledb.DB_TYPE_TIMESTAMP_TZ,
                None,
                None,
                None,
            ),
            (
                "TIMESTAMPLTZVALUE",
                oracledb.DB_TYPE_TIMESTAMP_LTZ,
                None,
                None,
                None,
            ),
            ("CLOBVALUE", oracledb.DB_TYPE_CLOB, None, None, None),
            ("NCLOBVALUE", oracledb.DB_TYPE_NCLOB, None, None, None),
            ("BLOBVALUE", oracledb.DB_TYPE_BLOB, None, None, None),
            ("SUBOBJECTVALUE", sub_obj_type, None, None, None),
            ("SUBOBJECTARRAY", sub_arr_type, None, None, None),
        ]
        obj_type = self.conn.gettype("TESTALLTYPES%ROWTYPE")
        actual_metadata = [
            (attr.name, attr.type, attr.precision, attr.scale, attr.max_size)
            for attr in obj_type.attributes
        ]
        self.assertEqual(actual_metadata, expected_metadata)

    def test_2338(self):
        "2338 - test collection iteration"
        self.cursor.execute("select udt_array(5, 10, 15) from dual")
        (obj,) = self.cursor.fetchone()
        result = [i for i in obj]
        self.assertEqual(result, [5, 10, 15])

    @unittest.skipUnless(
        test_env.get_is_thin(), "thick mode does not support xmltype"
    )
    def test_2339(self):
        "2339 - test fetching an object containing an XmlType instance"
        num_val = 2339
        xml_val = "<item>test_2339</item>"
        str_val = "A string for test 2339"
        self.cursor.execute(
            f"""
            select udt_ObjectWithXmlType({num_val}, sys.xmltype('{xml_val}'),
                    '{str_val}') from dual
            """
        )
        (obj,) = self.cursor.fetchone()
        self.assertEqual(obj.NUMBERVALUE, num_val)
        self.assertEqual(obj.XMLVALUE, xml_val)
        self.assertEqual(obj.STRINGVALUE, str_val)

    def test_2340(self):
        "2340 - test DbObject instances are retained across getvalue() calls"
        typ = self.conn.gettype("UDT_OBJECT")
        obj = typ.newobject()
        var = self.cursor.var(typ)
        var.setvalue(0, obj)
        self.assertIs(var.getvalue(), obj)

    def test_2341(self):
        "2341 - test insufficient privileges for gettype()"
        user = test_env.get_proxy_user()
        password = test_env.get_proxy_password()
        main_user = test_env.get_main_user().upper()
        conn = test_env.get_connection(user=user, password=password)
        with self.assertRaisesFullCode("DPY-2035"):
            conn.gettype(f"{main_user}.UDT_OBJECTARRAY")

    def test_2342(self):
        "2342 - test nested records"
        options = [(None, None), (1, None), (None, 2), (1, 2)]
        typ = self.conn.gettype("PKG_TESTNESTEDRECORDS.UDT_OUTER")
        for option in options:
            with self.subTest(option=option):
                value1, value2 = option
                obj = self.cursor.callfunc(
                    "pkg_TestNestedRecords.GetOuter", typ, (value1, value2)
                )
                self.assertIsNotNone(obj.INNER1)
                self.assertIsNone(obj.INNER1.ATTR1)
                self.assertEqual(obj.INNER1.ATTR2, value1)
                self.assertIsNotNone(obj.INNER2)
                self.assertIsNone(obj.INNER2.ATTR1)
                self.assertEqual(obj.INNER2.ATTR2, value2)


if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# aq_array.pyx
#
# Cython file defining the messages that are sent to the database and the
# responses that are received by the client for enqueuing and dequeuing
# an array of AQ messages (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

@cython.final
cdef class AqArrayMessage(AqBaseMessage):
    cdef:
        list props_impls
        int operation
        uint32_t num_iters

    cdef int _initialize_hook(self) except -1:
        """
        Perform initialization
        """
        self.function_code = TNS_FUNC_ARRAY_AQ

    cdef int _process_return_parameters(self, ReadBuffer buf) except -1:
        """
        Process the return parameters of the AQ array enqueue/dequeue request.
        """
        cdef:
            uint32_t i, j, num_iters, temp32
            ThinMsgPropsImpl props_impl
            uint16_t temp16
            bytes msgid
        buf.read_ub4(&num_iters)
        for i in range(num_iters):
            props_impl = self.props_impls[i]
            buf.read_ub2(&temp16)
            if temp16 > 0:
                buf.skip_ub1()
                self._process_msg_props(buf, props_impl)
            self._process_recipients(buf)
            buf.read_ub2(&temp16)
            if temp16 > 0:
                props_impl.payload = self._process_payload(buf)
            msgid = buf.read_bytes_with_length()
            if self.operation == TNS_AQ_ARRAY_ENQ:
                for j, props_impl in enumerate(self.props_impls):
                    props_impl.msgid = msgid[j * 16:(j + 1) * 16]
            else:
                props_impl.msgid = msgid
            buf.read_ub2(&temp16)               # extensions len
            if temp16 > 0:
                errors._raise_err(errors.ERR_NOT_IMPLEMENTED)
            buf.skip_ub2()                      # output ack
        if self.operation == TNS_AQ_ARRAY_ENQ:
            buf.read_ub4(&self.num_iters)
        else:
            self.num_iters = num_iters

    cdef int _write_array_deq(self, WriteBuffer buf) except -1:
        """
        Writes to the buffer the fields specific to the array dqeueue of AQ
        messages.
        """
        cdef:
            bytes consumer_name_bytes = None
            bytes correlation_bytes = None
            bytes condition_bytes = None
            ThinMsgPropsImpl props_impl
            bytes queue_name_bytes
            uint16_t delivery_mode
            uint32_t flags = 0

        # setup
        queue_name_bytes = self.queue_impl.name.encode()
        delivery_mode = self.deq_options_impl.delivery_mode
        if delivery_mode == TNS_AQ_MSG_BUFFERED:
            flags |= TNS_KPD_AQ_BUFMSG
        elif delivery_mode == TNS_AQ_MSG_PERSISTENT_OR_BUFFERED:
            flags |= TNS_KPD_AQ_EITHER
        if self.deq_options_impl.consumer_name:
            consumer_name_bytes = self.deq_options_impl.consumer_name.encode()
        if self.deq_options_impl.condition:
            condition_bytes = self.deq_options_impl.condition.encode()
        if self.deq_options_impl.correlation:
            correlation_bytes = self.deq_options_impl.correlation.encode()

        # write message
        for props_impl in self.props_impls:
            buf.write_ub4(len(queue_name_bytes))
            buf.write_bytes_with_length(queue_name_bytes)
            self._write_msg_props(buf, props_impl)
            buf.write_ub4(0)                        # num recipients
            self._write_value_with_length(buf, consumer_name_bytes)
            buf.write_sb4(self.deq_options_impl.mode)
            buf.write_sb4(self.deq_options_impl.navigation)
            buf.write_sb4(self.deq_options_impl.visibility)
            buf.write_sb4(self.deq_options_impl.wait)
            self._write_value_with_length(buf, self.deq_options_impl.msgid)
            self._write_value_with_length(buf, correlation_bytes)
            self._write_value_with_length(buf, condition_bytes)
            buf.write_ub4(0)                        # extensions
            buf.write_ub4(0)                        # rel msg id
            buf.write_sb4(0)                        # seq deviation
            buf.write_ub4(16)                       # toid length
            buf.write_bytes_with_length(self.queue_impl.payload_toid)
            buf.write_ub2(TNS_AQ_MESSAGE_VERSION)
            buf.write_ub4(0)                        # payload length
            buf.write_ub4(0)                        # raw pay length
            buf.write_ub4(0)
            buf.write_ub4(flags)
            buf.write_ub4(0)                        # extensions len
            buf.write_ub4(0)                        # source seq len

    cdef int _write_array_enq(self, WriteBuffer buf) except -1:
        """
        Writing input parameters incase of array enqueue
        """
        cdef:
            ThinMsgPropsImpl props_impl
            bytes queue_name_bytes
            uint32_t flags = 0

        # setup
        queue_name_bytes = self.queue_impl.name.encode()
        if self.enq_options_impl.delivery_mode == TNS_AQ_MSG_BUFFERED:
            flags |= TNS_KPD_AQ_BUFMSG

        # write message
        buf.write_ub4(0)                            # rel msgid len
        buf.write_uint8(TNS_MSG_TYPE_ROW_HEADER)
        buf.write_ub4(len(queue_name_bytes))
        buf.write_bytes_with_length(queue_name_bytes)
        buf.write_bytes(self.queue_impl.payload_toid)
        buf.write_ub2(TNS_AQ_MESSAGE_VERSION)
        buf.write_ub4(flags)
        for props_impl in self.props_impls:
            buf.write_uint8(TNS_MSG_TYPE_ROW_DATA)
            buf.write_ub4(flags)                    # aqi flags
            self._write_msg_props(buf, props_impl)
            buf.write_ub4(0)                        # num recipients
            buf.write_sb4(self.enq_options_impl.visibility)
            buf.write_ub4(0)                        # relative msg id
            buf.write_sb4(0)                        # seq deviation
            if self.queue_impl.payload_type is None \
                    and not self.queue_impl.is_json:
                buf.write_ub4(len(props_impl.payload_obj))
            self._write_payload(buf, props_impl)
        buf.write_uint8(TNS_MSG_TYPE_STATUS)

    cdef int _write_message(self, WriteBuffer buf) except -1:
        """
        Write message to the network buffers.
        """
        self._write_function_code(buf)
        if self.operation == TNS_AQ_ARRAY_ENQ:
            buf.write_uint8(0)                      # input params
            buf.write_ub4(0)                        # length
        else:
            buf.write_uint8(1)
            buf.write_ub4(self.num_iters)
        buf.write_ub4(TNS_AQ_ARRAY_FLAGS_RETURN_MESSAGE_ID)
        if self.operation == TNS_AQ_ARRAY_ENQ:
            buf.write_uint8(1)                      # output params
            buf.write_uint8(0)                      # length
        else:
            buf.write_uint8(1)
            buf.write_uint8(1)
        buf.write_sb4(self.operation)
        if self.operation == TNS_AQ_ARRAY_ENQ:
            buf.write_uint8(1)                      # num iters (pointer)
        else:
            buf.write_uint8(0)
        if buf._caps.ttc_field_version >= TNS_CCAP_FIELD_VERSION_21_1:
            buf.write_ub4(0xffff)                   # shard id
        if self.operation == TNS_AQ_ARRAY_ENQ:
            buf.write_ub4(self.num_iters)
        if self.operation == TNS_AQ_ARRAY_ENQ:
            self._write_array_enq(buf)
        else:
            self._write_array_deq(buf)



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# var.pyx
#
# Cython file defining the base Variable implementation class (embedded in
# base_impl.pyx).
#------------------------------------------------------------------------------

@cython.freelist(20)
cdef class BaseVarImpl:

    cdef int _bind(self, object conn, BaseCursorImpl cursor,
                   uint32_t num_execs, object name, uint32_t pos) except -1:
        """
        Binds a variable to the cursor.
        """
        raise NotImplementedError()


    cdef int _check_and_set_scalar_value(self, uint32_t pos, object value,
                                         bint* was_set) except -1:
        """
        Sets a scalar value in the variable at the given position, but first
        checks the type of Python value to see if it is acceptable. The value
        may be modified by the in converter (if one has been set) or adjusted
        to be acceptable (for some cases). If the was_set pointer is NULL, an
        exception is raised when the Python value is found to be unacceptable;
        otherwise, the flag is cleared if the Python value is unacceptable.
        """
        cdef uint32_t size

        # call in converter, if applicable
        if self.inconverter is not None:
            value = self.inconverter(value)

        # check the value and verify it is acceptable
        value = self._conn_impl._check_value(self.metadata, value, was_set)
        if was_set != NULL and not was_set[0]:
            return 0

        # resize variable, if applicable
        if value is not None and self.metadata.dbtype.default_size != 0:
            size = <uint32_t> len(value)
            if size > self.metadata.max_size:
                self._resize(size)

        # set value
        self._set_scalar_value(pos, value)
        self._is_value_set = True

    cdef int _check_and_set_value(self, uint32_t pos, object value,
                                  bint* was_set) except -1:
        """
        Sets the value in the variable at the given position, but first checks
        the type of Python value to see if it is acceptable.
        """
        cdef:
            uint32_t i, num_elements_in_array
            object element_value

        # scalar variables can be checked directly
        if not self.is_array:
            return self._check_and_set_scalar_value(pos, value, was_set)

        # array variables must have a list supplied to them
        if not isinstance(value, list):
            if was_set != NULL:
                was_set[0] = False
                return 0
            errors._raise_err(errors.ERR_EXPECTING_LIST_FOR_ARRAY_VAR)

        # the size of the array must be sufficient to hold all of the
        # elements
        num_elements_in_array = len(<list> value)
        if num_elements_in_array > self.num_elements:
            if was_set != NULL:
                was_set[0] = False
                return 0
            errors._raise_err(errors.ERR_INCORRECT_VAR_ARRAYSIZE,
                              var_arraysize=self.num_elements,
                              required_arraysize=num_elements_in_array)

        # check and set each of the element's values
        for i, element_value in enumerate(<list> value):
            self._check_and_set_scalar_value(i, element_value, was_set)
            if was_set != NULL and not was_set[0]:
                return 0
        self._set_num_elements_in_array(num_elements_in_array)

    cdef DbType _check_fetch_conversion(self):
        """
        Checks to see if the fetch type can be transformed into the requested
        type. This is only called during fetch when an output type handler is
        specified and the supplied variable has a type other than the fetch
        type. All matches are done using the underlying Oracle type and the
        character set form of the variable metadata is adjusted to
        use the character set form of the fetch metadata where applicable. The
        type to use for the metadata is returned.
        """
        cdef uint8_t from_ora_type_num, to_ora_type_num
        from_ora_type_num = self._fetch_metadata.dbtype._ora_type_num
        to_ora_type_num = self.metadata.dbtype._ora_type_num

        # Oracle BINARY_DOUBLE, BINARY_FLOAT
        if from_ora_type_num in (ORA_TYPE_NUM_BINARY_DOUBLE,
                                 ORA_TYPE_NUM_BINARY_FLOAT):
            if to_ora_type_num == ORA_TYPE_NUM_BINARY_INTEGER:
                self.metadata._py_type_num = PY_TYPE_NUM_INT
                return self._get_adjusted_type(to_ora_type_num)
            elif to_ora_type_num in (ORA_TYPE_NUM_BINARY_DOUBLE,
                                     ORA_TYPE_NUM_BINARY_FLOAT,
                                     ORA_TYPE_NUM_NUMBER):
                self.metadata._py_type_num = PY_TYPE_NUM_FLOAT
                return self._get_adjusted_type(to_ora_type_num)
            elif to_ora_type_num in (ORA_TYPE_NUM_CHAR,
                                     ORA_TYPE_NUM_LONG,
                                     ORA_TYPE_NUM_VARCHAR):
                return self._get_adjusted_type(to_ora_type_num)

        # Oracle BINARY_INTEGER
        elif from_ora_type_num == ORA_TYPE_NUM_BINARY_INTEGER:
            if to_ora_type_num == ORA_TYPE_NUM_NUMBER:
                self.metadata._py_type_num = PY_TYPE_NUM_FLOAT
                return self._get_adjusted_type(to_ora_type_num)
            elif to_ora_type_num in (ORA_TYPE_NUM_CHAR,
                                     ORA_TYPE_NUM_LONG,
                                     ORA_TYPE_NUM_VARCHAR):
                return self._get_adjusted_type(to_ora_type_num)

        # Oracle BLOB
        elif from_ora_type_num == ORA_TYPE_NUM_BLOB:
            if to_ora_type_num in (ORA_TYPE_NUM_RAW, ORA_TYPE_NUM_LONG_RAW):
                self._fetch_metadata.dbtype = DB_TYPE_LONG_RAW
                return self._fetch_metadata.dbtype

        # Oracle CHAR, LONG or VARCHAR
        elif from_ora_type_num in (ORA_TYPE_NUM_CHAR,
                                   ORA_TYPE_NUM_LONG,
                                   ORA_TYPE_NUM_VARCHAR):
            if to_ora_type_num in (ORA_TYPE_NUM_BINARY_DOUBLE,
                                   ORA_TYPE_NUM_BINARY_FLOAT,
                                   ORA_TYPE_NUM_NUMBER):
                self.metadata._py_type_num = PY_TYPE_NUM_FLOAT
                return self._get_adjusted_type(to_ora_type_num)
            elif to_ora_type_num == ORA_TYPE_NUM_BINARY_INTEGER:
                self.metadata._py_type_num = PY_TYPE_NUM_INT
                return self._get_adjusted_type(to_ora_type_num)
            elif to_ora_type_num in (ORA_TYPE_NUM_CHAR,
                                     ORA_TYPE_NUM_LONG,
                                     ORA_TYPE_NUM_VARCHAR):
                return self._get_adjusted_type(to_ora_type_num)

        # Oracle CLOB
        elif from_ora_type_num == ORA_TYPE_NUM_CLOB:
            if to_ora_type_num in (ORA_TYPE_NUM_CHAR,
                                   ORA_TYPE_NUM_LONG,
                                   ORA_TYPE_NUM_VARCHAR):
                self._fetch_metadata.dbtype = \
                        self._get_adjusted_type(ORA_TYPE_NUM_LONG)
                return self._fetch_metadata.dbtype

        # Oracle DATE, TIMESTAMP (WITH (LOCAL) TIME ZONE)
        elif from_ora_type_num in (ORA_TYPE_NUM_DATE,
                                   ORA_TYPE_NUM_TIMESTAMP,
                                   ORA_TYPE_NUM_TIMESTAMP_LTZ,
                                   ORA_TYPE_NUM_TIMESTAMP_TZ):
            if to_ora_type_num in (ORA_TYPE_NUM_DATE,
                                   ORA_TYPE_NUM_TIMESTAMP,
                                   ORA_TYPE_NUM_TIMESTAMP_LTZ,
                                   ORA_TYPE_NUM_TIMESTAMP_TZ,
                                   ORA_TYPE_NUM_CHAR,
                                   ORA_TYPE_NUM_LONG,
                                   ORA_TYPE_NUM_VARCHAR):
                return self._get_adjusted_type(to_ora_type_num)

        # Oracle INTERVAL DAY TO SECOND, INTERVAL YEAR TO MONTH, ROWID
        elif from_ora_type_num in (ORA_TYPE_NUM_INTERVAL_DS,
                                   ORA_TYPE_NUM_INTERVAL_YM,
                                   ORA_TYPE_NUM_ROWID):
            if to_ora_type_num in (ORA_TYPE_NUM_CHAR,
                                   ORA_TYPE_NUM_LONG,
                                   ORA_TYPE_NUM_VARCHAR):
                return self._get_adjusted_type(to_ora_type_num)

        # Oracle JSON
        elif from_ora_type_num == ORA_TYPE_NUM_JSON:
            if to_ora_type_num in (ORA_TYPE_NUM_CHAR,
                                   ORA_TYPE_NUM_VARCHAR):
                # the server won't accept LONG being defined but even so it
                # still sends back LONG data!
                self._fetch_metadata.dbtype = DB_TYPE_LONG
                return DB_TYPE_VARCHAR
            elif to_ora_type_num == ORA_TYPE_NUM_RAW:
                self._fetch_metadata.dbtype = DB_TYPE_RAW
                return self._fetch_metadata.dbtype

        # Oracle NUMBER
        elif from_ora_type_num == ORA_TYPE_NUM_NUMBER:
            if to_ora_type_num == ORA_TYPE_NUM_BINARY_INTEGER:
                self.metadata._py_type_num = PY_TYPE_NUM_INT
                return self._get_adjusted_type(to_ora_type_num)
            elif to_ora_type_num in (ORA_TYPE_NUM_BINARY_DOUBLE,
                                     ORA_TYPE_NUM_BINARY_FLOAT):
                self.metadata._py_type_num = PY_TYPE_NUM_FLOAT
                return self._get_adjusted_type(to_ora_type_num)
            elif to_ora_type_num in (ORA_TYPE_NUM_CHAR,
                                     ORA_TYPE_NUM_LONG,
                                     ORA_TYPE_NUM_VARCHAR):
                return self._get_adjusted_type(to_ora_type_num)

        # Oracle VECTOR
        elif from_ora_type_num == ORA_TYPE_NUM_VECTOR:
            if to_ora_type_num in (ORA_TYPE_NUM_CHAR,
                                   ORA_TYPE_NUM_LONG,
                                   ORA_TYPE_NUM_VARCHAR):
                self._fetch_metadata.dbtype = DB_TYPE_LONG
                return self._fetch_metadata.dbtype
            elif to_ora_type_num == ORA_TYPE_NUM_CLOB:
                self._fetch_metadata.dbtype = DB_TYPE_CLOB
                return self._fetch_metadata.dbtype

        # conversion not supported
        errors._raise_err(errors.ERR_INCONSISTENT_DATATYPES,
                          input_type=self._fetch_metadata.dbtype.name,
                          output_type=self.metadata.dbtype.name)

    cdef int _create_arrow_array(self) except -1:
        """
        Creates an Arrow array based on the type information selected by the
        user.
        """
        cdef ArrowTimeUnit time_unit = NANOARROW_TIME_UNIT_SECOND
        self.metadata._set_arrow_type()
        if self.metadata._arrow_type == NANOARROW_TYPE_TIMESTAMP:
            if self.metadata.scale > 0 and self.metadata.scale <= 3:
                time_unit = NANOARROW_TIME_UNIT_MILLI
            elif self.metadata.scale > 3 and self.metadata.scale <= 6:
                time_unit = NANOARROW_TIME_UNIT_MICRO
            elif self.metadata.scale > 6 and self.metadata.scale <= 9:
                time_unit = NANOARROW_TIME_UNIT_NANO
        self._arrow_array = OracleArrowArray(
            arrow_type=self.metadata._arrow_type,
            name=self.metadata.name,
            precision=self.metadata.precision,
            scale=self.metadata.scale,
            time_unit=time_unit,
        )

    cdef int _finalize_init(self) except -1:
        """
        Internal method that finalizes initialization of the variable.
        """
        self.metadata._finalize_init()
        if self.num_elements == 0:
            self.num_elements = 1
        self._has_returned_data = False

    cdef OracleArrowArray _finish_building_arrow_array(self):
        """
        Finish building the Arrow array associated with the variable and then
        return that array (after clearing it in the variable so that a new
        array will be built if more rows are fetched).
        """
        cdef OracleArrowArray array = self._arrow_array
        array.finish_building()
        self._arrow_array = None
        return array

    cdef DbType _get_adjusted_type(self, uint8_t ora_type_num):
        """
        Returns an adjusted type based on the desired Oracle type and the
        character set form from the fetch metadata. The intent of this function
        is to ensure that the character set form is not changed. For example,
        if the user requests the use of NVARCHAR but the data being fetched is
        of type CLOB, the type will be adjusted to VARCHAR. For those types
        which do not require a character set form, CS_FORM_IMPLICIT is always
        used.
        """
        cdef uint8_t csfrm
        if self.metadata.dbtype._csfrm == 0:
            csfrm = 0
        elif self._fetch_metadata.dbtype._csfrm == 0:
            csfrm = CS_FORM_IMPLICIT
        else:
            csfrm = self._fetch_metadata.dbtype._csfrm
        return DbType._from_ora_type_and_csfrm(ora_type_num, csfrm)

    cdef list _get_array_value(self):
        """
        Internal method to return the value of the array.
        """
        raise NotImplementedError()

    cdef object _get_scalar_value(self, uint32_t pos):
        """
        Internal method to return the value of the variable at the given
        position.
        """
        raise NotImplementedError()

    cdef int _on_reset_bind(self, uint32_t num_rows) except -1:
        """
        Called when the bind variable is being reset, just prior to performing
        a bind operation.
        """
        if self.num_elements < num_rows:
            self.num_elements = num_rows
            self._finalize_init()

    cdef int _resize(self, uint32_t new_size) except -1:
        """
        Resize the variable to the new size provided.
        """
        self.metadata.max_size = new_size
        self.metadata.buffer_size = 0
        self.metadata._finalize_init()

    cdef int _set_metadata_from_type(self, object typ) except -1:
        """
        Sets the type and size of the variable given a Python type.
        """
        self.metadata = OracleMetadata.from_type(typ)

    cdef int _set_metadata_from_value(self, object value,
                                      bint is_plsql) except -1:
        """
        Sets the type and size of the variable given a Python value. This
        method is called once for scalars and once per element in a list for
        array values. If a different type is detected an error is raised.
        """
        cdef OracleMetadata metadata
        metadata = OracleMetadata.from_value(value)
        if metadata.dbtype is DB_TYPE_BOOLEAN \
                and not self._conn_impl.supports_bool and not is_plsql:
            metadata.dbtype = DB_TYPE_BINARY_INTEGER
        if self.metadata is None:
            self.metadata = metadata
        elif metadata.dbtype is not self.metadata.dbtype \
                or metadata.objtype is not self.metadata.objtype:
            errors._raise_err(errors.ERR_MIXED_ELEMENT_TYPES, element=value)
        elif metadata.max_size > self.metadata.max_size:
            self.metadata.max_size = metadata.max_size

    cdef int _set_num_elements_in_array(self, uint32_t num_elements) except -1:
        """
        Sets the number of elements in the array.
        """
        self.num_elements_in_array = num_elements

    cdef int _set_scalar_value(self, uint32_t pos, object value) except -1:
        """
        Set the value of the variable at the given position. At this point it
        is assumed that all checks have been performed!
        """
        raise NotImplementedError()

    def get_all_values(self):
        """
        Internal method for returning an array of all of the values stored in
        the variable.
        """
        cdef uint32_t i
        if self.is_array:
            return self._get_array_value()
        return [self._get_scalar_value(i) for i in range(self.num_elements)]

    def get_value(self, uint32_t pos):
        """
        Internal method for getting the value of a variable.
        """
        if self.is_array:
            return self._get_array_value()
        if pos >= self.num_elements:
            raise IndexError("position out of range")
        return self._get_scalar_value(pos)

    def set_value(self, uint32_t pos, object value):
        """
        Internal method for setting a variable's value at the specified
        position.
        """
        if self.is_array:
            if pos > 0:
                errors._raise_err(errors.ERR_ARRAYS_OF_ARRAYS)
        elif pos >= self.num_elements:
            raise IndexError("position out of range")
        self._check_and_set_value(pos, value, NULL)



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# auth.pyx
#
# Cython file defining the messages sent to the database and the responses that
# are received by the client for authentication (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

@cython.final
cdef class AuthMessage(Message):
    cdef:
        str encoded_password
        bytes password
        bytes newpassword
        str encoded_newpassword
        str encoded_jdwp_data
        str debug_jdwp
        str session_key
        str speedy_key
        str proxy_user
        str token
        str private_key
        str service_name
        uint8_t purity
        ssize_t user_bytes_len
        bytes user_bytes
        dict session_data
        uint32_t auth_mode
        uint32_t verifier_type
        bint change_password
        str program
        str terminal
        str machine
        str osuser
        str driver_name
        str edition
        list appcontext
        str connect_string

    cdef int _encrypt_passwords(self) except -1:
        """
        Encrypts the passwords using the session key.
        """

        # encrypt password
        salt = secrets.token_bytes(16)
        password_with_salt = salt + self.password
        encrypted_password = encrypt_cbc(self.conn_impl._combo_key,
                                         password_with_salt)
        self.encoded_password = encrypted_password.hex().upper()

        # encrypt new password
        if self.newpassword is not None:
            newpassword_with_salt = salt + self.newpassword
            encrypted_newpassword = encrypt_cbc(self.conn_impl._combo_key,
                                                newpassword_with_salt)
            self.encoded_newpassword = encrypted_newpassword.hex().upper()

    cdef int _generate_verifier(self) except -1:
        """
        Generate the multi-round verifier.
        """
        cdef:
            bytes jdwp_data
            bytearray b
            ssize_t i

        # create password hash
        verifier_data = bytes.fromhex(self.session_data['AUTH_VFR_DATA'])
        if self.verifier_type == TNS_VERIFIER_TYPE_12C:
            keylen = 32
            iterations = int(self.session_data['AUTH_PBKDF2_VGEN_COUNT'])
            salt = verifier_data + b'AUTH_PBKDF2_SPEEDY_KEY'
            password_key = get_derived_key(self.password, salt, 64,
                                           iterations)
            h = hashlib.new("sha512")
            h.update(password_key)
            h.update(verifier_data)
            password_hash = h.digest()[:32]
        else:
            keylen = 24
            h = hashlib.sha1(self.password)
            h.update(verifier_data)
            password_hash = h.digest() + bytes(4)

        # decrypt first half of session key
        encoded_server_key = bytes.fromhex(self.session_data['AUTH_SESSKEY'])
        session_key_part_a = decrypt_cbc(password_hash, encoded_server_key)

        # generate second half of session key
        session_key_part_b = secrets.token_bytes(len(session_key_part_a))
        encoded_client_key = encrypt_cbc(password_hash, session_key_part_b)

        # create session key and combo key
        if len(session_key_part_a) == 48:
            self.session_key = encoded_client_key.hex().upper()[:96]
            b = bytearray(24)
            for i in range(16, 40):
                b[i - 16] = session_key_part_a[i] ^ session_key_part_b[i]
            part1 = hashlib.md5(b[:16]).digest()
            part2 = hashlib.md5(b[16:]).digest()
            combo_key = (part1 + part2)[:keylen]
        else:
            self.session_key = encoded_client_key.hex().upper()[:64]
            salt = bytes.fromhex(self.session_data['AUTH_PBKDF2_CSK_SALT'])
            iterations = int(self.session_data['AUTH_PBKDF2_SDER_COUNT'])
            temp_key = session_key_part_b[:keylen] + session_key_part_a[:keylen]
            combo_key = get_derived_key(temp_key.hex().upper().encode(), salt,
                                        keylen, iterations)

        # retain session key for use by the change password API
        self.conn_impl._combo_key = combo_key

        # generate speedy key for 12c verifiers
        if self.verifier_type == TNS_VERIFIER_TYPE_12C:
            salt = secrets.token_bytes(16)
            speedy_key = encrypt_cbc(combo_key, salt + password_key)
            self.speedy_key = speedy_key[:80].hex().upper()

        # encrypts the passwords
        self._encrypt_passwords()

        # check if debug_jdwp is set. if set, encode the data using the
        # combo session key with zeros padding
        if self.debug_jdwp is not None:
            jdwp_data = self.debug_jdwp.encode()
            encrypted_jdwp_data = encrypt_cbc(combo_key, jdwp_data, zeros=True)
            # Add a "01" at the end of the hex encrypted data to indicate the
            # use of AES encryption
            self.encoded_jdwp_data = encrypted_jdwp_data.hex().upper() + "01"

    cdef str _get_alter_timezone_statement(self):
        """
        Returns the statement required to change the session time zone to match
        the time zone in use by the Python interpreter.
        """
        cdef:
            int tz_hour, tz_minute, timezone
            str sign, tz_repr
        tz_repr = os.environ.get("ORA_SDTZ")
        if tz_repr is None:
            timezone = time.localtime().tm_gmtoff
            tz_hour = timezone // 3600
            tz_minute = (timezone - (tz_hour * 3600)) // 60
            if tz_hour < 0:
                sign = "-"
                tz_hour = -tz_hour
            else:
                sign = "+"
            tz_repr = f"{sign}{tz_hour:02}:{tz_minute:02}"
        return f"ALTER SESSION SET TIME_ZONE='{tz_repr}'\x00"

    cdef tuple _get_version_tuple(self, ReadBuffer buf):
        """
        Return the 5-tuple for the database version. Note that the format
        changed with Oracle Database 18.
        """
        cdef uint32_t full_version_num
        full_version_num = int(self.session_data["AUTH_VERSION_NO"])
        if buf._caps.ttc_field_version >= TNS_CCAP_FIELD_VERSION_18_1_EXT_1:
            return ((full_version_num >> 24) & 0xFF,
                    (full_version_num >> 16) & 0xFF,
                    (full_version_num >> 12) & 0x0F,
                    (full_version_num >> 4) & 0xFF,
                    (full_version_num & 0x0F))
        else:
            return ((full_version_num >> 24) & 0xFF,
                    (full_version_num >> 20) & 0x0F,
                    (full_version_num >> 12) & 0x0F,
                    (full_version_num >> 8) & 0x0F,
                    (full_version_num & 0x0F))

    cdef int _initialize_hook(self) except -1:
        """
        Perform initialization.
        """
        self.function_code = TNS_FUNC_AUTH_PHASE_ONE
        self.session_data = {}
        if self.conn_impl.username is not None:
            self.user_bytes = self.conn_impl.username.encode()
            self.user_bytes_len = len(self.user_bytes)
        self.resend = True

    cdef int _process_return_parameters(self, ReadBuffer buf) except -1:
        cdef:
            uint16_t num_params, i
            str key, value
        buf.read_ub2(&num_params)
        for i in range(num_params):
            key = buf.read_str_with_length()
            value = buf.read_str_with_length()
            if value is None:
                value = ""
            if key == "AUTH_VFR_DATA":
                buf.read_ub4(&self.verifier_type)
            else:
                buf.skip_ub4()                  # skip flags
            self.session_data[key] = value
        if self.function_code == TNS_FUNC_AUTH_PHASE_ONE:
            self.function_code = TNS_FUNC_AUTH_PHASE_TWO
        elif not self.change_password:
            self.conn_impl._session_id = \
                    <uint32_t> int(self.session_data["AUTH_SESSION_ID"])
            self.conn_impl._serial_num = \
                    <uint16_t> int(self.session_data["AUTH_SERIAL_NUM"])
            self.conn_impl._db_domain = \
                    self.session_data.get("AUTH_SC_DB_DOMAIN")
            self.conn_impl._db_name = \
                    self.session_data.get("AUTH_SC_DBUNIQUE_NAME")
            self.conn_impl._max_open_cursors = \
                    int(self.session_data.get("AUTH_MAX_OPEN_CURSORS", 0))
            self.conn_impl._service_name = \
                    self.session_data.get("AUTH_SC_SERVICE_NAME")
            self.conn_impl._instance_name = \
                    self.session_data.get("AUTH_INSTANCENAME")
            self.conn_impl._max_identifier_length = \
                    int(self.session_data.get("AUTH_MAX_IDEN_LENGTH", 30))
            self.conn_impl.server_version = self._get_version_tuple(buf)
            self.conn_impl.supports_bool = \
                    buf._caps.ttc_field_version >= TNS_CCAP_FIELD_VERSION_23_1
            self.conn_impl._edition = self.edition

    cdef int _set_params(self, ConnectParamsImpl params,
                         Description description) except -1:
        """
        Sets the parameters to use for the AuthMessage. The user and auth mode
        are retained in order to avoid duplicating this effort for both trips
        to the server.
        """
        self.password = params._get_password()
        self.newpassword = params._get_new_password()
        self.service_name = description.service_name
        self.proxy_user = params.proxy_user
        self.debug_jdwp = params.debug_jdwp
        self.program = params.program
        self.terminal = params.terminal
        self.machine = params.machine
        self.osuser = params.osuser
        self.driver_name = params.driver_name
        if self.driver_name is None:
            self.driver_name = f"{DRIVER_NAME} thn : {DRIVER_VERSION}"
        self.edition = params.edition
        self.appcontext = params.appcontext
        self.connect_string = params._get_connect_string()

        # if drcp is used, use purity = NEW as the default purity for
        # standalone connections and purity = SELF for connections that belong
        # to a pool
        if description.purity == PURITY_DEFAULT \
                and self.conn_impl._drcp_enabled:
            if self.conn_impl._pool is None:
                self.purity = PURITY_NEW
            else:
                self.purity = PURITY_SELF
        else:
            self.purity = description.purity

        # set token parameters; adjust processing so that only phase two is
        # sent
        if params._token is not None \
                or params.access_token_callback is not None:
            self.token = params._get_token()
            if params._private_key is not None:
                self.private_key = params._get_private_key()
            self.function_code = TNS_FUNC_AUTH_PHASE_TWO
            self.resend = False

        # set authentication mode
        if params._new_password is None:
            self.auth_mode = TNS_AUTH_MODE_LOGON
        if params.mode & AUTH_MODE_SYSDBA:
            self.auth_mode |= TNS_AUTH_MODE_SYSDBA
        if params.mode & AUTH_MODE_SYSOPER:
            self.auth_mode |= TNS_AUTH_MODE_SYSOPER
        if params.mode & AUTH_MODE_SYSASM:
            self.auth_mode |= TNS_AUTH_MODE_SYSASM
        if params.mode & AUTH_MODE_SYSBKP:
            self.auth_mode |= TNS_AUTH_MODE_SYSBKP
        if params.mode & AUTH_MODE_SYSDGD:
            self.auth_mode |= TNS_AUTH_MODE_SYSDGD
        if params.mode & AUTH_MODE_SYSKMT:
            self.auth_mode |= TNS_AUTH_MODE_SYSKMT
        if params.mode & AUTH_MODE_SYSRAC:
            self.auth_mode |= TNS_AUTH_MODE_SYSRAC
        if self.private_key is not None:
            self.auth_mode |= TNS_AUTH_MODE_IAM_TOKEN

    cdef int _write_key_value(self, WriteBuffer buf, str key, str value,
                              uint32_t flags=0) except -1:
        cdef:
            bytes key_bytes = key.encode()
            bytes value_bytes = value.encode()
            uint32_t key_len = <uint32_t> len(key_bytes)
            uint32_t value_len = <uint32_t> len(value_bytes)
        buf.write_ub4(key_len)
        buf.write_bytes_with_length(key_bytes)
        buf.write_ub4(value_len)
        if value_len > 0:
            buf.write_bytes_with_length(value_bytes)
        buf.write_ub4(flags)

    cdef int _write_message(self, WriteBuffer buf) except -1:
        cdef:
            uint8_t has_user = 1 if self.user_bytes_len > 0 else 0
            uint32_t num_pairs

        # perform final determination of data to write
        if self.function_code == TNS_FUNC_AUTH_PHASE_ONE:
            num_pairs = 5
        elif self.change_password:
            self._encrypt_passwords()
            num_pairs = 2
        else:
            num_pairs = 4

            # token authentication
            if self.token is not None:
                num_pairs += 1

            # normal user/password authentication
            else:
                num_pairs += 2
                self.auth_mode |= TNS_AUTH_MODE_WITH_PASSWORD
                if self.verifier_type == TNS_VERIFIER_TYPE_12C:
                    num_pairs += 1
                elif self.verifier_type not in (TNS_VERIFIER_TYPE_11G_1,
                                                TNS_VERIFIER_TYPE_11G_2):
                    errors._raise_err(errors.ERR_UNSUPPORTED_VERIFIER_TYPE,
                                      verifier_type=self.verifier_type)
                self._generate_verifier()

            # determine which other key/value pairs to write
            if self.newpassword is not None:
                num_pairs += 1
                self.auth_mode |= TNS_AUTH_MODE_CHANGE_PASSWORD
            if self.proxy_user is not None:
                num_pairs += 1
            if self.conn_impl._cclass is not None:
                num_pairs += 1
            if self.purity != 0:
                num_pairs += 1
            if self.private_key is not None:
                num_pairs += 2
            if self.encoded_jdwp_data is not None:
                num_pairs += 1
            if self.edition is not None:
                num_pairs += 1
            if self.appcontext is not None:
                num_pairs += len(self.appcontext) * 3
            if self.connect_string is not None:
                num_pairs += 1

        # write basic data to packet
        self._write_function_code(buf)
        buf.write_uint8(has_user)           # pointer (authusr)
        buf.write_ub4(self.user_bytes_len)
        buf.write_ub4(self.auth_mode)       # authentication mode
        buf.write_uint8(1)                  # pointer (authivl)
        buf.write_ub4(num_pairs)            # number of key/value pairs
        buf.write_uint8(1)                  # pointer (authovl)
        buf.write_uint8(1)                  # pointer (authovln)
        if has_user:
            buf.write_bytes_with_length(self.user_bytes)

        # write key/value pairs
        if self.function_code == TNS_FUNC_AUTH_PHASE_ONE:
            self._write_key_value(buf, "AUTH_TERMINAL", self.terminal)
            self._write_key_value(buf, "AUTH_PROGRAM_NM", self.program)
            self._write_key_value(buf, "AUTH_MACHINE", self.machine)
            self._write_key_value(buf, "AUTH_PID", _connect_constants.pid)
            self._write_key_value(buf, "AUTH_SID", self.osuser)
        else:
            if self.proxy_user is not None:
                self._write_key_value(buf, "PROXY_CLIENT_NAME",
                                      self.proxy_user)
            if self.token is not None:
                self._write_key_value(buf, "AUTH_TOKEN", self.token)
            elif not self.change_password:
                self._write_key_value(buf, "AUTH_SESSKEY", self.session_key, 1)
                if self.verifier_type == TNS_VERIFIER_TYPE_12C:
                    self._write_key_value(buf, "AUTH_PBKDF2_SPEEDY_KEY",
                                          self.speedy_key)
            if self.encoded_password is not None:
                self._write_key_value(buf, "AUTH_PASSWORD",
                                      self.encoded_password)
            if self.encoded_newpassword is not None:
                self._write_key_value(buf, "AUTH_NEWPASSWORD",
                                      self.encoded_newpassword)
            if not self.change_password:
                self._write_key_value(buf, "SESSION_CLIENT_CHARSET", "873")
                self._write_key_value(buf, "SESSION_CLIENT_DRIVER_NAME",
                                      self.driver_name)
                self._write_key_value(buf, "SESSION_CLIENT_VERSION",
                                    str(_connect_constants.full_version_num))
                self._write_key_value(buf, "AUTH_ALTER_SESSION",
                                      self._get_alter_timezone_statement(), 1)
            if self.conn_impl._cclass is not None:
                self._write_key_value(buf, "AUTH_KPPL_CONN_CLASS",
                                      self.conn_impl._cclass)
            if self.purity != 0:
                self._write_key_value(buf, "AUTH_KPPL_PURITY",
                                      str(self.purity), 1)
            if self.private_key is not None:
                date_format = "%a, %d %b %Y %H:%M:%S GMT"
                now = datetime.datetime.utcnow().strftime(date_format)
                host_info = "%s:%d" % buf._transport.get_host_info()
                header = f"date: {now}\n" + \
                         f"(request-target): {self.service_name}\n" + \
                         f"host: {host_info}"
                signature = get_signature(self.private_key, header)
                self._write_key_value(buf, "AUTH_HEADER", header)
                self._write_key_value(buf, "AUTH_SIGNATURE", signature)
            if self.encoded_jdwp_data is not None:
                self._write_key_value(buf, "AUTH_ORA_DEBUG_JDWP",
                                      self.encoded_jdwp_data)
            if self.edition is not None:
                self._write_key_value(buf, "AUTH_ORA_EDITION", self.edition)
            if self.appcontext is not None:
                # NOTE: these keys require a trailing null character as the
                # server expects it!
                for entry in self.appcontext:
                    self._write_key_value(buf, "AUTH_APPCTX_NSPACE\0", entry[0])
                    self._write_key_value(buf, "AUTH_APPCTX_ATTR\0", entry[1])
                    self._write_key_value(buf, "AUTH_APPCTX_VALUE\0", entry[2])
            if self.connect_string is not None:
                self._write_key_value(buf, "AUTH_CONNECT_STRING",
                                      self.connect_string)



# -----------------------------------------------------------------------------
# Copyright (c) 2020, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
1900 - Module for testing LOB (CLOB and BLOB) variables
"""

import pickle
import unittest

import oracledb
import test_env


class TestCase(test_env.BaseTestCase):
    def __get_temp_lobs(self, sid):
        cursor = self.conn.cursor()
        cursor.execute(
            """
            select cache_lobs + nocache_lobs + abstract_lobs
            from v$temporary_lobs
            where sid = :sid
            """,
            sid=sid,
        )
        row = cursor.fetchone()
        if row is None:
            return 0
        return int(row[0])

    def __perform_test(self, lob_type, input_type):
        long_string = ""
        db_type = getattr(oracledb, f"DB_TYPE_{lob_type}")
        self.cursor.execute(f"delete from Test{lob_type}s")
        self.conn.commit()
        for i in range(11):
            if i > 0:
                char = chr(ord("A") + i - 1)
                long_string += char * 25000
            elif input_type is not db_type:
                continue
            self.cursor.setinputsizes(long_string=input_type)
            if lob_type == "BLOB":
                bind_value = long_string.encode()
            else:
                bind_value = long_string
            self.cursor.execute(
                f"""
                insert into Test{lob_type}s (IntCol, {lob_type}Col)
                values (:integer_value, :long_string)
                """,
                integer_value=i,
                long_string=bind_value,
            )
        self.conn.commit()
        self.cursor.execute(
            f"""
            select IntCol, {lob_type}Col
            from Test{lob_type}s
            order by IntCol
            """
        )
        self.__validate_query(self.cursor, lob_type)

    def __test_bind_ordering(self, lob_type):
        main_col = "A" * 32768
        extra_col_1 = "B" * 65536
        extra_col_2 = "C" * 131072
        if lob_type == "BLOB":
            main_col = main_col.encode()
            extra_col_1 = extra_col_1.encode()
            extra_col_2 = extra_col_2.encode()
        self.conn.stmtcachesize = 0
        self.cursor.execute(f"delete from Test{lob_type}s")
        self.conn.commit()
        data = (1, main_col, 8, extra_col_1, 15, extra_col_2)
        self.cursor.execute(
            f"""
            insert into Test{lob_type}s (IntCol, {lob_type}Col,
                ExtraNumCol1, Extra{lob_type}Col1, ExtraNumCol2,
                Extra{lob_type}Col2)
            values (:1, :2, :3, :4, :5, :6)
            """,
            data,
        )
        with test_env.DefaultsContextManager("fetch_lobs", False):
            self.cursor.execute(f"select * from Test{lob_type}s")
            self.assertEqual(self.cursor.fetchone(), data)

    def __test_fetch_lobs_direct(self, lob_type):
        self.cursor.execute(f"delete from Test{lob_type}s")
        self.conn.commit()
        data = []
        long_string = ""
        for i in range(1, 11):
            if i > 0:
                char = chr(ord("A") + i - 1)
                long_string += char * 25000
            if lob_type == "BLOB":
                data.append((i, long_string.encode()))
            else:
                data.append((i, long_string))
        self.cursor.executemany(
            f"""
            insert into Test{lob_type}s (IntCol, {lob_type}Col)
            values (:1, :2)
            """,
            data,
        )
        with test_env.DefaultsContextManager("fetch_lobs", False):
            self.cursor.execute(
                f"""
                select IntCol, {lob_type}Col
                from Test{lob_type}s
                order by IntCol
                """
            )
            self.assertEqual(self.cursor.fetchall(), data)

    def __test_lob_operations(self, lob_type):
        self.cursor.execute(f"delete from Test{lob_type}s")
        self.conn.commit()
        self.cursor.setinputsizes(long_string=getattr(oracledb, lob_type))
        long_string = "X" * 75000
        write_value = "TEST"
        if lob_type == "BLOB":
            long_string = long_string.encode("ascii")
            write_value = write_value.encode("ascii")
        self.cursor.execute(
            f"""
            insert into Test{lob_type}s (IntCol, {lob_type}Col)
            values (:integer_value, :long_string)
            """,
            integer_value=1,
            long_string=long_string,
        )
        self.cursor.execute(
            f"""
            select {lob_type}Col
            from Test{lob_type}s
            where IntCol = 1
            """
        )
        (lob,) = self.cursor.fetchone()
        self.assertFalse(lob.isopen())
        lob.open()
        with self.assertRaisesFullCode("ORA-22293"):
            lob.open()
        self.assertTrue(lob.isopen())
        lob.close()
        with self.assertRaisesFullCode("ORA-22289"):
            lob.close()
        self.assertFalse(lob.isopen())
        self.assertEqual(lob.size(), 75000)
        lob.write(write_value, 75001)
        self.assertRaises(TypeError, lob.write, 1000, 1)
        self.assertRaises(TypeError, lob.write, "data", "1")
        self.assertEqual(lob.size(), 75000 + len(write_value))
        with self.assertRaisesFullCode("DPY-2030"):
            lob.read(0)
        with self.assertRaisesFullCode("DPY-2030"):
            lob.read(-25)
        with self.assertRaisesFullCode("DPY-2047"):
            lob.read(amount=0)
        with self.assertRaisesFullCode("DPY-2047"):
            lob.read(amount=-5)
        self.assertEqual(lob.read(), long_string + write_value)
        lob.write(write_value, 1)
        self.assertEqual(
            lob.read(), write_value + long_string[4:] + write_value
        )
        lob.trim(25000)
        self.assertEqual(lob.size(), 25000)
        lob.trim(newSize=10000)
        self.assertEqual(lob.size(), 10000)
        with self.assertRaisesFullCode("DPY-2014"):
            lob.trim(new_size=50, newSize=60)
        self.assertRaises(TypeError, lob.trim, new_size="10000")
        self.assertRaises(TypeError, lob.trim, newSize="10000")
        lob.trim(new_size=40)
        self.assertEqual(lob.size(), 40)
        lob.trim()
        self.assertEqual(lob.size(), 0)
        self.assertIsInstance(lob.getchunksize(), int)

    def __test_pickle(self, lob_type):
        value = "A test string value for pickling"
        if lob_type == "BLOB":
            value = value.encode("ascii")
        db_type = getattr(oracledb, "DB_TYPE_" + lob_type)
        lob = self.conn.createlob(db_type, value)
        pickled_data = pickle.dumps(lob)
        unpickled_value = pickle.loads(pickled_data)
        self.assertEqual(unpickled_value, value)

    def __test_temporary_lob(self, lob_type):
        self.cursor.execute(f"delete from Test{lob_type}s")
        self.conn.commit()
        value = "A test string value"
        if lob_type == "BLOB":
            value = value.encode("ascii")
        db_type = getattr(oracledb, f"DB_TYPE_{lob_type}")
        lob = self.conn.createlob(db_type, value)
        self.cursor.execute(
            f"""
            insert into Test{lob_type}s (IntCol, {lob_type}Col)
            values (:int_val, :lob_val)
            """,
            int_val=1,
            lob_val=lob,
        )
        self.cursor.execute(f"select {lob_type}Col from Test{lob_type}s")
        (lob,) = self.cursor.fetchone()
        self.assertEqual(lob.read(), value)

    def __validate_query(self, rows, lob_type):
        long_string = ""
        db_type = getattr(oracledb, f"DB_TYPE_{lob_type}")
        for row in rows:
            integer_value, lob = row
            self.assertEqual(lob.type, db_type)
            if integer_value == 0:
                self.assertEqual(lob.size(), 0)
                expected_value = ""
                if lob_type == "BLOB":
                    expected_value = expected_value.encode()
                self.assertEqual(lob.read(), expected_value)
            else:
                char = chr(ord("A") + integer_value - 1)
                prev_char = chr(ord("A") + integer_value - 2)
                long_string += char * 25000
                if lob_type == "BLOB":
                    expected_value = long_string.encode("ascii")
                    char = char.encode("ascii")
                    prev_char = prev_char.encode("ascii")
                else:
                    expected_value = long_string
                self.assertEqual(lob.size(), len(expected_value))
                self.assertEqual(lob.read(), expected_value)
                if lob_type == "CLOB":
                    self.assertEqual(str(lob), expected_value)
                self.assertEqual(lob.read(len(expected_value)), char)
            if integer_value > 1:
                offset = (integer_value - 1) * 25000 - 4
                string = prev_char * 5 + char * 5
                self.assertEqual(lob.read(offset, 10), string)

    def test_1900(self):
        "1900 - test binding a LOB value directly"
        self.cursor.execute("delete from TestCLOBs")
        self.cursor.execute(
            """
            insert into TestCLOBs
            (IntCol, ClobCol)
            values (1, 'Short value')
            """
        )
        self.conn.commit()
        self.cursor.execute("select ClobCol from TestCLOBs")
        (lob,) = self.cursor.fetchone()
        self.cursor.execute(
            """
            insert into TestCLOBs
            (IntCol, ClobCol)
            values (2, :value)
            """,
            value=lob,
        )

    def test_1901(self):
        "1901 - test cursor description is accurate for BLOBs"
        self.cursor.execute("select IntCol, BlobCol from TestBLOBs")
        expected_value = [
            ("INTCOL", oracledb.DB_TYPE_NUMBER, 10, None, 9, 0, 0),
            ("BLOBCOL", oracledb.DB_TYPE_BLOB, None, None, None, None, 0),
        ]
        self.assertEqual(self.cursor.description, expected_value)

    def test_1902(self):
        "1902 - test binding and fetching BLOB data (directly)"
        self.__perform_test("BLOB", oracledb.DB_TYPE_BLOB)

    def test_1903(self):
        "1903 - test binding and fetching BLOB data (indirectly)"
        self.__perform_test("BLOB", oracledb.DB_TYPE_LONG_RAW)

    def test_1904(self):
        "1904 - test operations on BLOBs"
        self.__test_lob_operations("BLOB")

    def test_1905(self):
        "1905 - test cursor description is accurate for CLOBs"
        self.cursor.execute("select IntCol, ClobCol from TestCLOBs")
        expected_value = [
            ("INTCOL", oracledb.DB_TYPE_NUMBER, 10, None, 9, 0, False),
            ("CLOBCOL", oracledb.DB_TYPE_CLOB, None, None, None, None, False),
        ]
        self.assertEqual(self.cursor.description, expected_value)

    def test_1906(self):
        "1906 - test binding and fetching CLOB data (directly)"
        self.__perform_test("CLOB", oracledb.DB_TYPE_CLOB)

    def test_1907(self):
        "1907 - test binding and fetching CLOB data (indirectly)"
        self.__perform_test("CLOB", oracledb.DB_TYPE_LONG)

    def test_1908(self):
        "1908 - test operations on CLOBs"
        self.__test_lob_operations("CLOB")

    def test_1909(self):
        "1909 - test creating a temporary BLOB"
        self.__test_temporary_lob("BLOB")

    def test_1910(self):
        "1910 - test creating a temporary CLOB"
        self.__test_temporary_lob("CLOB")

    def test_1911(self):
        "1911 - test creating a temporary NCLOB"
        self.__test_temporary_lob("NCLOB")

    def test_1912(self):
        "1912 - test retrieving data from a CLOB after multiple fetches"
        self.cursor.arraysize = 1
        self.__perform_test("CLOB", oracledb.DB_TYPE_CLOB)

    def test_1913(self):
        "1913 - test cursor description is accurate for NCLOBs"
        self.cursor.execute("select IntCol, NClobCol from TestNCLOBs")
        expected_value = [
            ("INTCOL", oracledb.DB_TYPE_NUMBER, 10, None, 9, 0, 0),
            ("NCLOBCOL", oracledb.DB_TYPE_NCLOB, None, None, None, None, 0),
        ]
        self.assertEqual(self.cursor.description, expected_value)

    def test_1914(self):
        "1914 - test binding and fetching NCLOB data (directly)"
        self.__perform_test("NCLOB", oracledb.DB_TYPE_NCLOB)

    def test_1915(self):
        "1915 - test binding and fetching NCLOB data (with non-ASCII chars)"
        value = "\u03b4\u4e2a"
        self.cursor.execute("delete from TestNCLOBs")
        self.cursor.setinputsizes(val=oracledb.DB_TYPE_NVARCHAR)
        self.cursor.execute(
            """
            insert into TestNCLOBs (IntCol, NClobCol)
            values (1, :val)
            """,
            val=value,
        )
        self.conn.commit()
        self.cursor.execute("select NCLOBCol from TestNCLOBs")
        (nclob,) = self.cursor.fetchone()
        self.cursor.setinputsizes(val=oracledb.DB_TYPE_NVARCHAR)
        self.cursor.execute(
            "update TestNCLOBs set NCLOBCol = :val", val=nclob.read() + value
        )
        self.cursor.execute("select NCLOBCol from TestNCLOBs")
        (nclob,) = self.cursor.fetchone()
        self.assertEqual(nclob.read(), value + value)

    def test_1916(self):
        "1916 - test binding and fetching NCLOB data (indirectly)"
        self.__perform_test("NCLOB", oracledb.DB_TYPE_LONG)

    def test_1917(self):
        "1917 - test operations on NCLOBs"
        self.__test_lob_operations("NCLOB")

    @unittest.skipIf(
        test_env.get_is_implicit_pooling(),
        "sessions can change with implicit pooling",
    )
    def test_1918(self):
        "1918 - test temporary LOBs"
        self.cursor.execute("select sys_context('USERENV', 'SID') from dual")
        (sid,) = self.cursor.fetchone()
        temp_lobs = self.__get_temp_lobs(sid)
        with self.conn.cursor() as cursor:
            cursor.arraysize = 27
            self.assertEqual(temp_lobs, 0)
            cursor.execute(
                "select extract(xmlcol, '/').getclobval() from TestXML"
            )
            for (lob,) in cursor:
                lob.read()
                del lob
        temp_lobs = self.__get_temp_lobs(sid)
        self.assertEqual(temp_lobs, 0)

    def test_1919(self):
        "1919 - test assign string to NCLOB beyond array size"
        nclobVar = self.cursor.var(oracledb.DB_TYPE_NCLOB)
        self.assertRaises(IndexError, nclobVar.setvalue, 1, "test char")

    def test_1920(self):
        "1920 - test read/write temporary LOBs using supplemental characters"
        if test_env.get_charset() != "AL32UTF8":
            self.skipTest("Database character set must be AL32UTF8")
        supplemental_chars = (
            "𠜎 𠜱 𠝹 𠱓 𠱸 𠲖 𠳏 𠳕 𠴕 𠵼 𠵿 𠸎 𠸏 𠹷 𠺝 𠺢 𠻗 𠻹 𠻺 𠼭 𠼮 "
            "𠽌 𠾴 𠾼 𠿪 𡁜 𡁯 𡁵 𡁶 𡁻 𡃁 𡃉 𡇙 𢃇 𢞵 𢫕 𢭃 𢯊 𢱑 𢱕 𢳂 𢴈 "
            "𢵌 𢵧 𢺳 𣲷 𤓓 𤶸 𤷪 𥄫 𦉘 𦟌 𦧲 𦧺 𧨾 𨅝 𨈇 𨋢 𨳊 𨳍 𨳒 𩶘"
        )
        self.cursor.execute("delete from TestCLOBs")
        lob = self.conn.createlob(oracledb.DB_TYPE_CLOB, supplemental_chars)
        self.cursor.execute(
            """
            insert into TestCLOBs
            (IntCol, ClobCol)
            values (1, :val)
            """,
            [lob],
        )
        self.conn.commit()
        self.cursor.execute("select ClobCol from TestCLOBs")
        (lob,) = self.cursor.fetchone()
        self.assertEqual(lob.read(), supplemental_chars)

    def test_1921(self):
        "1921 - test automatic conversion to CLOB for PL/SQL"
        var = self.cursor.var(str, outconverter=lambda v: v[-15:])
        var.setvalue(0, "A" * 50000)
        self.cursor.execute(
            """
            declare
                t_Clob          clob;
            begin
                t_Clob := :data;
                dbms_lob.copy(:data, t_Clob, 50000);
                dbms_lob.writeappend(:data, 5, 'BBBBB');
            end;
            """,
            data=var,
        )
        self.assertEqual(var.getvalue(), "A" * 10 + "B" * 5)

    def test_1922(self):
        "1922 - test automatic conversion to NCLOB for PL/SQL"
        var = self.cursor.var(
            oracledb.DB_TYPE_NCHAR, outconverter=lambda v: v[-12:]
        )
        var.setvalue(0, "N" * 51234)
        self.cursor.execute(
            """
            declare
                t_Clob          nclob;
            begin
                t_Clob := :data;
                dbms_lob.copy(:data, t_Clob, 51234);
                dbms_lob.writeappend(:data, 7, 'PPPPPPP');
            end;
            """,
            data=var,
        )
        self.assertEqual(var.getvalue(), "N" * 5 + "P" * 7)

    def test_1923(self):
        "1923 - test automatic conversion to BLOB for PL/SQL"
        var = self.cursor.var(bytes, outconverter=lambda v: v[-14:])
        var.setvalue(0, b"L" * 52345)
        self.cursor.execute(
            """
            declare
                t_Blob          blob;
            begin
                t_Blob := :data;
                dbms_lob.copy(:data, t_Blob, 52345);
                dbms_lob.writeappend(:data, 6, '515151515151');
            end;
            """,
            data=var,
        )
        self.assertEqual(var.getvalue(), b"L" * 8 + b"Q" * 6)

    def test_1924(self):
        "1924 - test pickling of BLOB"
        self.__test_pickle("BLOB")

    def test_1925(self):
        "1925 - test pickling of CLOB"
        self.__test_pickle("CLOB")

    def test_1926(self):
        "1925 - test pickling of NCLOB"
        self.__test_pickle("NCLOB")

    def test_1927(self):
        "1927 - test fetching BLOB as bytes"
        self.__test_fetch_lobs_direct("BLOB")

    def test_1928(self):
        "1928 - test fetching CLOB as str"
        self.__test_fetch_lobs_direct("CLOB")

    def test_1929(self):
        "1929 - test fetching NCLOB as str"
        self.__test_fetch_lobs_direct("NCLOB")

    def test_1930(self):
        "1930 - test bind ordering with BLOB"
        self.__test_bind_ordering("BLOB")

    def test_1931(self):
        "1931 - test bind ordering with CLOB"
        self.__test_bind_ordering("CLOB")

    def test_1932(self):
        "1932 - test bind ordering with NCLOB"
        self.__test_bind_ordering("NCLOB")

    def test_1933(self):
        "1933 - test creating a lob with an invalid type"
        with self.assertRaises(TypeError):
            self.conn.createlob(oracledb.DB_TYPE_NUMBER)
        with self.assertRaises(TypeError):
            self.conn.createlob(oracledb.DB_TYPE_BFILE)

    def test_1934(self):
        "1934 - test creation of temporary LOBs with varying data"
        cases = [
            (oracledb.DB_TYPE_BLOB, b"test_1934A", b"!", b"test_1934A!"),
            (oracledb.DB_TYPE_BLOB, "test_1934B", "!", b"test_1934B!"),
            (oracledb.DB_TYPE_CLOB, b"test_1934C", b"!", "test_1934C!"),
            (oracledb.DB_TYPE_CLOB, "test_1934D", "!", "test_1934D!"),
            (oracledb.DB_TYPE_NCLOB, b"test_1934E", b"!", "test_1934E!"),
            (oracledb.DB_TYPE_NCLOB, "test_1934F", "!", "test_1934F!"),
        ]
        for typ, initial_data, additional_data, expected_result in cases:
            with self.subTest():
                lob = self.conn.createlob(typ, initial_data)
                lob.write(additional_data, len(initial_data) + 1)
                self.assertEqual(lob.read(), expected_result)

    def test_1935(self):
        "1935 - test reading and writing a LOB with a closed connection"
        types = [
            oracledb.DB_TYPE_BLOB,
            oracledb.DB_TYPE_CLOB,
            oracledb.DB_TYPE_NCLOB,
        ]
        for typ in types:
            conn = test_env.get_connection()
            lob = conn.createlob(typ, "Temp LOB")
            conn.close()
            with self.assertRaisesFullCode("DPY-1001"):
                lob.read()
            with self.assertRaisesFullCode("DPY-1001"):
                lob.write("x")

    def test_1936(self):
        "1936 - test reading a non-existent directory"
        directory_name = "TEST_1936_MISSING_DIR"
        file_name = "test_1936_missing_file.txt"
        self.cursor.execute(
            "select BFILENAME(:1, :2) from dual", [directory_name, file_name]
        )
        (bfile,) = self.cursor.fetchone()
        self.assertEqual(bfile.getfilename(), (directory_name, file_name))
        with self.assertRaisesFullCode("ORA-22285"):
            bfile.fileexists()
        with self.assertRaisesFullCode("ORA-22285"):
            bfile.read()

    def test_1937(self):
        "1937 - test using BFILE methods on non-BFILE LOBs"
        types = [
            oracledb.DB_TYPE_BLOB,
            oracledb.DB_TYPE_CLOB,
            oracledb.DB_TYPE_NCLOB,
        ]
        for typ in types:
            lob = self.conn.createlob(typ)
            with self.assertRaisesFullCode("DPY-3026"):
                lob.getfilename()
            with self.assertRaisesFullCode("DPY-3026"):
                lob.setfilename("not_relevant", "not_relevant")
            with self.assertRaisesFullCode("DPY-3026"):
                lob.fileexists()

    def test_1938(self):
        "1938 - confirm that LOB objects are retained across getvalue() calls"
        for typ in (
            oracledb.DB_TYPE_BLOB,
            oracledb.DB_TYPE_CLOB,
            oracledb.DB_TYPE_NCLOB,
        ):
            var = self.cursor.var(typ)
            lob = self.conn.createlob(typ, "Some data for test 1938")
            var.setvalue(0, lob)
            self.assertIs(var.getvalue(), lob)

    def test_1939(self):
        "1939 - temporary LOB in/out without modification"
        value = "test - 1939"
        var = self.cursor.var(oracledb.DB_TYPE_CLOB)
        var.setvalue(0, value)
        self.assertEqual(var.getvalue().read(), value)
        self.cursor.callproc("pkg_TestLOBs.TestInOut", [var, None, None])
        self.assertEqual(var.getvalue().read(), value)

    def test_1940(self):
        "1940 - temporary LOB in/out with modification"
        search_value = "test"
        replace_value = "replaced"
        initial_value = f"{search_value} - 1939"
        final_value = f"{replace_value} - 1939"
        var = self.cursor.var(oracledb.DB_TYPE_CLOB)
        var.setvalue(0, initial_value)
        self.assertEqual(var.getvalue().read(), initial_value)
        self.cursor.callproc(
            "pkg_TestLOBs.TestInOut", [var, search_value, replace_value]
        )
        self.assertEqual(var.getvalue().read(), final_value)


if __name__ == "__main__":
    test_env.run_test_cases()
